# -*- coding: utf-8 -*-
# functions.py

# Copyright (C) 2008-2010 concentricpuddle, GPLv2

"""A modules that defines function that are to be used througout puddletag.

Basically, this modules contains functions that are called via
the getfunc function in findfunc.py. Most of the time though,
getfunc isn't called directly, but via the tagtofilename function
also found in the findfunc module.

The docstrings of these functions are also used as metadata in the
actiondlg modules in order to draw windows to allow the user
to create and edit actiondlg.Function objects.

The first parameter of a function should be name one of two things.
If it's named text, a line of text[like tag value] will be passed to it.
If it's named tags, then a dictionary of tags is passed to it.

The other restriction is that if the function is not successful, return None.

Now for the docstrings of the functions:
The first line of the docstring contains the function name as shown to
the user and the description(what the user sees once they've set values).

After the first line, there should be a line for each parameter of the function
not including the first[it's either tags or text].

actiondlg.FunctionDialog creates controls using this line.
This line is further split into three parts
    The first contains the Explanatory Label above the control.
    The Second contains the control itself, either text, combo or check
    The third contains the default arguments as shown to the user."""

import decimal
import math
import os
import re
import traceback
from mutagen.mp3 import HeaderNotFoundError
from collections import defaultdict
from functools import partial
from unidecode import unidecode

import pyparsing

from . import audioinfo
from .audioinfo import encode_fn
from .puddleobjects import (safe_name, fnmatch, natural_sort_key)

PATH = audioinfo.PATH
DIRPATH = audioinfo.DIRPATH

true = '1'
false = '0'
path = os.path

D = decimal.Decimal


def add(text, text1):
    try:
        return str((D(str(text)) + D(str(text1))).normalize().to_eng_string())
    except decimal.InvalidOperation:
        return


_padding = '0'


def _pad(text, numlen):
    if len(text) < numlen:
        text = _padding * ((numlen - len(text)) // len(_padding)) + text
    return text


def autonumbering(r_tags, minimum=1, restart=False, padding=1, state=None):
    '''Autonumbering, "Autonumbering: $0, Start: $1, Restart for dir: $2, Padding: $3"
oi,spinbox,1
aoeu,check, False
au,spinbox,1'''
    if restart:
        if 'autonumbering' not in state:
            state['autonumbering'] = defaultdict(lambda: 0)

        dircount = state['autonumbering']

        counter = dircount.get(r_tags.dirpath, 0) + 1
        dircount[r_tags.dirpath] = counter
    else:
        counter = int(state.get('__counter', 1))

    tracknum = str(minimum + counter - 1)

    if padding > 1:
        return _pad(tracknum, padding)
    else:
        return tracknum


def check_truth(text):
    if isinstance(text, str):
        text = text.strip()
    return 0 if ((not text) or (text == '0')) else 1


def and_(text, text1):
    return str(check_truth(text) and check_truth(text1))


def caps(text):
    # Capitalizes the first letter of each word in string and
    # converts the rest to lower case.
    return titleCase(text)


def caps2(text):
    # Capitalizes the first letter of each word in string and
    # leaves all other characters unchanged.
    upcase = set(i for i, char in enumerate(text) if char.upper() == char)
    return "".join(ch.upper() if i in upcase else ch
                   for i, ch in enumerate(text.title()))


def caps3(text):
    # Capitalizes the first letter of the string and converts
    # the rest to lower case.
    try:
        start = re.search(r"\w", text, re.U).start(0)
    except AttributeError:
        return
    return text[:start] + text[start].upper() + text[start + 1:].lower()


def ceiling(n_value):
    return math.ceil(n_value)


def char(text):
    try:
        return str(ord(text))
    except TypeError:
        return


def changeartist(artist, *files):
    for audio in files:
        audio['artist'] = artist
        audio.save()


def div(n_numerator, n_divisor):
    if n_divisor == 0:
        raise FuncError("Cannot divide by zero.")
    try:
        return str((D(n_numerator) / D(n_divisor)).normalize())
    except decimal.InvalidOperation:
        return
    # ret = unicode(float(n_numerator) / n_divisor)
    # if len(ret) < len(normalized):
    # return ret
    # else:
    # return normalized


def eql(text, text1):
    return true if text == text1 else false


# Contributed by Stjujsckij Nickolaj
def enconvert(text, enc_name):
    ''' Convert from non-standard encoding, "Convert to encoding: $0, Encoding: $1"
&Encoding, combo, cp1250, cp1251, cp1252, cp1253, cp1254, cp1255, cp1256, cp1257, cp1258,\
euc_jp, cp932, euc_jis_2004, shift_jis, johab, big5, big5hkscs, gb2312, gb18030, gbk, hz'''
    return text.encode("latin1", 'replace').decode(enc_name, 'replace')


def filenametotag(m_tags, p_pattern):
    """Filename to Tag, File->Tag '$1'
&Pattern, text"""
    return findfunc.filenametotag(p_pattern, m_tags[PATH], True)


def finddups(tracks, key='title', method=None):
    from .puddleobjects import dupes
    li = []
    for z in tracks:
        try:
            li.append(z[key])
        except KeyError:
            li.append(None)
    return dupes(li, method)


def floor(n_value):
    return math.floor(n_value)


def formatValue(m_tags, p_pattern, state=None):
    """Format value, Format $0 using $1
&Format string, text"""
    ret = findfunc.parsefunc(p_pattern, m_tags, state=state)
    if not ret:
        return
    else:
        return ret


format_value = formatValue


def geql(text, text1):
    try:
        text = float(text)
    except (TypeError, ValueError):
        pass

    try:
        text1 = float(text1)
    except (TypeError, ValueError):
        pass

    if text >= text1:
        return true
    else:
        return false


def grtr(text, text1):
    try:
        text = float(text) if text.strip() else 0
    except (TypeError, ValueError):
        pass

    try:
        text1 = float(text1) if text1.strip() else 0
    except (TypeError, ValueError):
        pass

    if text > text1:
        return true
    else:
        return false


def to_num(text):
    match = re.search(r'[\-\+]?[0-9]+(\.[0-9]+)?', text)
    return match.group() if match else ''


def hasformat(p_pat, tagname="__filename"):
    if findfunc.filenametotag(p_pat, tagname):
        return true
    return false


def if_(text, text1, z):
    if check_truth(text):
        return text1
    else:
        return z


def iflonger(a, b, text, text1):
    try:
        if len(a) > len(b):
            return text
        else:
            return text1
    except TypeError:
        return


def import_text(m_tags, p_pattern, r_tags):
    '''Import text file, "Text File: $0, '$1'"
&Pattern (can be relative path), text, lyrics.txt'''
    filename = tag_to_filename(p_pattern, m_tags, r_tags, False)
    if not filename:
        return
    try:
        with open(filename, 'tr', encoding='utf-8') as textfile:
            return textfile.read()
    except EnvironmentError:
        return
    except UnicodeDecodeError:
        return


def isdigit(text):
    try:
        decimal.Decimal(text)
        return true
    except decimal.InvalidOperation:
        return false


def left(text, n):
    try:
        n = int(n)
    except (TypeError, ValueError):
        raise FuncError('Integer expected, got "%s"' % str(n))
    return text[:n]


def len_(text):
    return str(len(text))


def leql(text, text1):
    try:
        text = float(text)
    except (TypeError, ValueError):
        pass

    try:
        text1 = float(text1)
    except (TypeError, ValueError):
        pass

    if text <= text1:
        return true
    else:
        return false


def less(text, text1):
    try:
        text = float(text)
    except (TypeError, ValueError):
        pass

    try:
        text1 = float(text1)
    except (TypeError, ValueError):
        pass

    if text < text1:
        return true
    else:
        return false


def libstuff(dirname):
    files = []
    for filename in os.listdir(dirname):
        tag = audioinfo.Tag(path.join(dirname, filename))
        if tag:
            files.append(tag)
    return files


def _load_image(filename):
    try:
        return {'data': open(filename, 'rb').read()}
    except EnvironmentError:
        traceback.print_exc()
        pass


def load_images(r_tags, filepatterns, desc, matchcase, state=None):
    '''Load Artwork, "Artwork: Filenames='$1', Description='$2', Case Sensitive=$3"
"&Filenames to check (;-separated, shell wildcards [eg. *] allowed)", text
&Default description (can be pattern):, text
Match filename's &case:, check'''
    tags = r_tags

    images = []
    dirpath = r_tags.dirpath
    pictures = fnmatch(filepatterns, os.listdir(dirpath), matchcase)
    for pic in pictures:
        filename = os.path.join(dirpath, pic)
        image = _load_image(filename)
        if not image:
            continue
        desc = formatValue(tags, desc)
        if desc is None:
            desc = ''
        image[audioinfo.DESCRIPTION] = desc
        image[audioinfo.IMAGETYPE] = 3
        images.append(image)

    if images:
        return {'__image': images}


def lower(text):
    return text.lower()


def merge_values(m_text, separator=';'):
    '''Merge field, "Merge field: $0, sep='$1'"
&Separator, text, ;'''
    if isinstance(m_text, str):
        return m_text
    else:
        return separator.join(m_text)


def meta_sep(m_tags, p_field, p_sep=', '):
    value = m_tags.get(p_field)
    if value is None:
        return None
    elif isinstance(value, str):
        return value
    else:
        return p_sep.join(value)


def meta(m_tags, field, n_index=None):
    value = m_tags.get(field)
    if value is None:
        return None
    if n_index is not None:
        n_index = int(n_index)
        try:
            return value[n_index]
        except IndexError:
            return ''
    else:
        if isinstance(value, str):
            return value
        return ', '.join(value)


def mid(text, n_start, n_len):
    try:
        n_start = int(n_start)
    except (TypeError, ValueError):
        raise FuncError('Integer expected, got "%s"' % str(n_start))

    try:
        n_len = int(n_len)
    except (TypeError, ValueError):
        raise FuncError('Integer expected, got "%s"' % str(n_len))

    return str(text)[n_start: n_start + n_len]


def mod(n_x, n_y):
    try:
        return str((n_x % n_y).normalize())
    except decimal.InvalidOperation:
        return


def tag_to_filename(pattern, m_tags, r_tags, ext=True, state=None,
                    is_dir=False):
    if not pattern:
        return
    if state is None:
        state = {}

    tags = m_tags
    tf = findfunc.parsefunc
    path_join = os.path.join

    if state is None:
        state = {}

    path_seps, text = tf(pattern, tags, state=state, path_sep="/")

    start_pos = 0
    new_dirs = []

    if path_seps:
        for p in path_seps:
            if start_pos != p:
                new_dirs.append(safe_name(text[start_pos:p]))
            start_pos = p
        new_dirs.append(safe_name(text[start_pos + 1:]))
    else:
        new_dirs = [safe_name(text)]

    if os.path.isabs(pattern):
        return add_extension('/' + '/'.join(map(safe_name, new_dirs)), tags, ext)
    else:

        subdirs = new_dirs
        count = len(path_seps)

        dirpath = r_tags.dirpath

        new_fn = encode_fn(path_join(*new_dirs))

        if new_fn.startswith('./'):
            return add_extension(path_join(dirpath, new_fn[len('./'):]), tags, ext)
        elif new_fn.startswith('../'):
            parent = dirpath
            while new_fn.startswith('../'):
                parent = os.path.dirname(parent)
                new_fn = new_fn[len('../'):]
            return add_extension(path_join(parent, new_fn), tags, ext)
        elif count and '..' not in subdirs:
            subdirs = dirpath.split('/')
            if count >= len(subdirs):
                parent = ['']
            else:
                if is_dir:
                    parent = subdirs[:-(count + 1)]
                else:
                    parent = subdirs[:-(count)]
        else:
            if is_dir:
                dirpath = os.path.dirname(r_tags.dirpath)
            parent = dirpath.split('/')

        if not parent[0]:
            parent.insert(0, '/')
        return add_extension(os.path.join(*(parent + [new_fn])), tags, ext)


def add_extension(fn, tags, addext=None, extension=None):
    if not addext:
        return fn
    elif addext and (extension is not None):
        return fn + os.path.extsep + encode_fn(extension)
    else:
        return fn + os.path.extsep + encode_fn(tags["__ext"])


def move(m_tags, p_pattern, r_tags, ext=True, state=None):
    """Tag to filename, Tag->File: $1
&Pattern, text"""

    tags = m_tags
    tf = findfunc.tagtofilename

    fn = tag_to_filename(p_pattern, m_tags, r_tags, ext, state)

    if fn:
        return {'__path': fn}


def mul(n_x, n_y):
    D = decimal.Decimal
    try:
        return str((D(n_x) * D(n_y)).normalize())
    except decimal.InvalidOperation:
        return


def neql(text, text1):
    if text == text1:
        return false
    else:
        return true


def not_(text):
    return false if check_truth(text) else true


def num(text, n_len, add_sep=false):
    if not text:
        return ""
    n_len = int(n_len)

    sep_index = text.find("/")

    if sep_index >= 0:
        total = text[sep_index + 1:]
        tracknum = text[:sep_index]
    else:
        total = None
        tracknum = text

    tracknum = tracknum.lstrip('0')

    if total and check_truth(add_sep):
        return "%s/%s" % (tracknum.zfill(n_len), total)
    else:
        return tracknum.zfill(n_len)


def odd(n_number):
    return true if (n_number % 2 != 0) else false


def or_(text, text1):
    return true if (check_truth(text) or check_truth(text1)) else false


def rand():
    import random
    return str(random.random())


def _round(n_value):
    return round(n_value)


def re_escape(rex, chars=r'^$[]\+*?.(){},|'):
    escaped = ""
    for ch in rex:
        if ch in chars:
            escaped = escaped + '\\' + ch
        else:
            escaped = escaped + ch
    return escaped


def pat_escape(p_pat):
    return re_escape(p_pat, '$%\\')


def remove_fields():
    '''Remove Fields, <blank> $0'''
    return ''


def remove_except(tags, fields):
    '''Remove all fields except, "Remove fields except: $1"
&Field list (; separated):, text, '''
    fields = [field for field in fields.split(';')]
    ret = dict([(field.strip(), '') for field in audioinfo.usertags(tags)
                if field not in fields])
    if '__image' not in fields:
        ret['__image'] = []
    if ret:
        return ret


import mutagen.id3, mutagen.apev2

_tag_classes = {
    'APEv2': mutagen.apev2.delete,
    'ID3v1': partial(mutagen.id3.delete, v1=True, v2=False),
    'ID3v2': partial(mutagen.id3.delete, v1=False, v2=True),
    'All ID3': partial(mutagen.id3.delete, v1=True, v2=True)}


def remove_tag(r_tags, tag='APEv2'):
    '''Remove Tag, "Remove $1 Tag"
&Tag, combo, Base, APEv2, ID3v1, ID3v2, All ID3'''

    if tag in _tag_classes:
        _tag_classes[tag](r_tags.filepath)


def rename_dirs(tags, state, pattern):
    '''Rename Directory, "Rename dir: $1"
&Pattern:, text'''
    dirname = safe_name(format_value(tags, pattern))
    old_path = tags['__dirpath']
    dirpath = path.join(path.dirname(old_path), dirname)
    if 'rename_dirs' in state:
        state['rename_dirs'][old_path] = dirpath
    else:
        state['rename_dirs'] = {old_path: dirpath}


def replace(text, word, replaceword, matchcase=False, whole=False, chars=None):
    '''Replace, "Replace $0: '$1' -> '$2', Match Case: $3, Words Only: $4"
&Replace, text
w&ith:, text
Match c&ase:, check
only as &whole word, check'''
    matchcase, whole = check_truth(matchcase), check_truth(whole)
    word = re_escape(word)

    if matchcase:
        matchcase = 0
    else:
        matchcase = re.IGNORECASE
    if chars is None:
        chars = r'\,\.\(\) \!\[\]'
    replaceword = replaceword.replace('\\', '\\\\')

    if whole:
        pat = re.compile(r'(^|[%s])%s([%s]|$)' % (chars, word, chars), matchcase)
    else:
        pat = re.compile(word, matchcase)

    start = 0
    match = pat.search(text, start)
    if whole:
        while match:
            start = match.start()
            end = match.end()
            sub = text[start: end]
            repl = replaceword
            if sub[0] in chars:
                repl = sub[0] + repl
            if sub[-1] in chars:
                repl = repl + sub[-1]

            text = pat.sub(repl, text, 1)
            match = pat.search(text, start + len(repl))
    else:
        try:
            text = pat.sub(replaceword, text)
        except Exception as e:
            raise findfunc.FuncError(str(e))
    return text


class RegHelper(object):
    def __init__(self, groups, repl):
        self.groups = groups
        self._repl = repl

    def repl(self, match):
        v = int(match.group()[1:])
        try:
            if re.search(r'\$[\w\d_]+\(', self._repl):
                return re_escape(self.groups[v], '"\\,')
            else:
                return self.groups[v]
        except KeyError:
            return '""'


def replaceWithReg(m_tags, text, regex, repl=None, matchcase=False, m_text=None, state=None):
    """Replace with RegExp, "RegReplace $0: RegExp '$1' with '$2', Match Case: $3"
&Regular Expression, text
Replace &matches with:, text
Match &Case, check"""

    if not regex:
        return text

    if m_text is None:
        m_text = [text]

    if not check_truth(matchcase):
        flags = re.UNICODE | re.I
    else:
        flags = re.UNICODE

    def replace_tokens(match):
        groups = match.groups()
        group = match.group()
        if not group:
            d = {}
        elif groups:
            d = dict([(i, z if z is not None else "") for i, z in enumerate(groups, 1)])
            d[0] = group
        else:
            d = {1: group, 0: group}

        ret = re.sub(r'(?i)\$\d+', RegHelper(d, repl).repl, repl, 0)
        return findfunc.parsefunc(ret, m_tags)

    def replace_matches(value):
        try:
            return re.sub(regex, replace_tokens, value, 0, flags)
        except re.error as e:
            raise findfunc.FuncError(str(e))

    return "\\".join(replace_matches(z) for z in m_text)


replace_regex = replaceWithReg


# Contributed by Erik Reckase
# Improved by David Gessel
def to_ascii(t_fn):
    """Converts all unicode chars to ASCII."""
    cleaned_fn = unidecode(t_fn, 'ignore')
    return ''.join(c for c in cleaned_fn if c.isprintable())


def remove_dupes(m_text, matchcase=False):
    """Remove duplicate values, "Remove Dupes: $0, Match Case $1"
Match &Case, check"""
    text = m_text
    if isinstance(text, str):
        return text

    if matchcase:
        ret = []
        append = ret.append
        [append(z) for z in text if z not in ret]
        return ret
    else:
        ret = []
        lowered = set()
        for z in text:
            if z.lower() not in lowered:
                lowered.add(z.lower())
                ret.append(z)
        return ret


def right(text, n):
    try:
        n = int(n)
    except (TypeError, ValueError):
        raise FuncError('Integer expected, got "%s"' % str(n))
    if n == 0:
        return ''
    return text[-int(n):]


def gain_to_watts(gain):
    return 10 ** (-gain * .1)


def to_hexstring(x):
    # leading space required; blame Apple
    return " %08X" % int(x)


def rg2sc(gain, peak=None):
    if peak is None:
        gain = gain.split(':')
        if len(gain) == 2:  # gain:peak
            peak = float(gain[1])
            gain = float(gain[0])
        elif len(gain) == 3:  # channel:gain:peak
            peak = float(gain[2])
            gain = float(gain[1])
        else:
            return
    else:
        gain = float(gain)
        peak = float(peak)

    values = [
        to_hexstring(1000 * gain_to_watts(gain)),
        to_hexstring(1000 * gain_to_watts(gain)),
        to_hexstring(2500 * gain_to_watts(gain)),
        to_hexstring(2500 * gain_to_watts(gain)),
        " 00024CA8",  # bogus
        " 00024CA8",  # bogus
        to_hexstring(peak * (32 * 1024 - 1)),
        to_hexstring(peak * (32 * 1024 - 1)),
        " 00024CA8",  # bogus
        " 00024CA8",  # bogus
    ]

    return str(''.join(values))


def save_artwork(m_tags, pattern, r_tags, state=None, write=True):
    """Export artwork to file, "Export Art: pattern='$1'"
&Pattern (extension not required), text, folder_%img_counter%"""
    if state is None:
        state = {}

    if 'artwork_data' not in state:
        state['artwork_data'] = set()

    if '__image' in m_tags:
        images = m_tags['__image']
    else:
        images = r_tags.images

    if not images:
        return

    new_state = state.copy()
    new_state['img_count'] = str(len(images))
    for i, image in enumerate(images):
        data = image[audioinfo.DATA]
        new_state['img_desc'] = image.get(audioinfo.DESCRIPTION, '')
        new_state['img_type'] = audioinfo.IMAGETYPES[
            image.get(audioinfo.IMAGETYPE, 3)]
        mime = image.get(audioinfo.MIMETYPE, '')
        if not mime:
            mime = audioinfo.get_mime(data)
            if not mime:
                continue

        extension = '.png' if 'png' in mime.lower() else '.jpg'

        new_state['img_mime'] = mime
        new_state['img_counter'] = str(i + 1)
        fn = tag_to_filename(pattern, m_tags, r_tags,
                             False, new_state) + extension

        if not fn:
            continue

        if data not in state['artwork_data']:
            state['artwork_data'].add(data)
        elif path.exists(fn):
            continue

        i = 1
        while path.exists(fn):
            fn = path.splitext(fn)[0] + '_' + str(i) + extension
            i += 1

        if write:
            fobj = open(fn, 'w+b')
            fobj.write(data)
            fobj.close()
        else:
            return fn


def sort_field(m_text, order='Ascending', matchcase=False):
    """Sort values, "Sort $0, order='$1', Match Case='$2'"
&Order, combo, Ascending, Descending,
Match &Case, check"""
    text = m_text

    if isinstance(text, str):
        return text
    return sorted(text,
                  key=lambda x: natural_sort_key(x, case_insensitive=not matchcase),
                  reverse=order != 'Ascending')


def split_by_sep(m_text, sep):
    """Split fields using separator, "Split using separator $0: sep='$1'"
&Separator, text, ;"""
    if isinstance(m_text, str):
        return m_text
    else:
        ret = []
        for t in m_text:
            try:
                ret.extend(t.split(sep))
            except ValueError:
                ret.append(t)
        return ret


def strip(text):
    '''Trim whitespace, Trim $0'''
    return text.strip()


def find(text, text1):
    val = text.find(text1)
    if val >= 0:
        return str(val)
    return '-1'


def sub(n_1, n_2):
    return str(n_1 - n_2)


def tag_dir(m_tags, pattern, r_tags, state=None):
    '''Tag to Dir, "Tag->Dir: $1"
&Pattern (can be relative path), text, %artist% - %album%'''
    if state is None:
        state = {'tag_dir': set()}

    elif 'tag_dir' not in state:
        state['tag_dir'] = set()

    if r_tags.dirpath in state['tag_dir']:
        return

    dirpath = r_tags.dirpath
    if pattern.endswith('/') and len(pattern) > 1:
        pattern = pattern[:-1]

    filename = tag_to_filename(pattern, m_tags, r_tags, False, state, True)
    if filename:
        state['tag_dir'].add(encode_fn(filename))
        return {DIRPATH: filename}


def testfunction(tags, t_text, p_pattern, n_number):
    text = '%s - %s' % (tags['artist'], tags['title'])
    assert t_text == text
    assert p_pattern == '%artist% - %title%'
    assert n_number == 23
    return 'Passed'


def texttotag(tags, input_text, p_pattern, output, state=None):
    """Text to Tag, "Text to Tag: $0 -> $1, $2"
&Text, text
&Pattern, text
&Output, text"""
    tagpattern = pyparsing.Literal('%').suppress() + \
                 pyparsing.Word(pyparsing.nums)
    input_text = findfunc.parsefunc(input_text, tags, state=state)
    d = findfunc.tagtotag(p_pattern, input_text, tagpattern)
    if d:
        for key in d:
            output = output.replace('%' +
                                    str(key), pat_escape(str(d[key])))
        return findfunc.parsefunc(output, tags, state=state)
    return None


def titleCase(text, ctype=None, characters=None):
    '''Case conversion, "Convert Case: $0: $1"
&Type, combo, Mixed Case,UPPER CASE,lower case
"For &Mixed Case, after any of:", text, "., !"'''
    if characters is None:
        characters = ['.', '(', ')', ' ', '!']
    if ctype == "UPPER CASE":
        return text.upper()
    elif ctype == 'lower case':
        return text.lower()

    text = [z for z in text]
    try:
        text[0] = text[0].upper()
    except IndexError:
        return ""
    for char in range(len(text)):
        try:
            if text[char] in characters:
                text[char + 1] = text[char + 1].upper()
            else:
                text[char + 1] = text[char + 1].lower()
        except IndexError:
            pass
    return "".join(text)


_update = {'APEv2': audioinfo.apev2.Tag, 'ID3': audioinfo.id3.Tag}


def update_from_tag(r_tags, fields, tag='APEv2'):
    '''Update from tag, "Update from $2, Fields: $1"
&Field list (; separated):, text,
&Tag, combo, APEv2, ID3'''
    try:
        tag = _update[tag]().link(r_tags.filepath)
        if tag is None:
            return
    except EnvironmentError:
        return
    except mutagen.mp3.HeaderNotFoundError:
        return
    fields = [_f for _f in [z.strip() for z in fields.split(';')] if _f]
    if not fields:
        return tag.usertags
    else:
        if fields[0].startswith('~'):
            return dict([(k, v) for k, v in tag.usertags.items()
                         if k not in fields])
        else:
            return dict([(k, v) for k, v in tag.usertags.items()
                         if k in fields])


def upper(text):
    return text.upper()


def validate(text, to=None, chars=None):
    from .puddleobjects import safe_name
    if chars is None:
        return safe_name(text, to=to)
    else:
        return safe_name(text, chars, to=to)


functions = {
    "add": add,
    "and": and_,
    'artwork': load_images,
    'autonumbering': autonumbering,
    "caps": caps,
    "caps2": caps2,
    "caps3": caps3,
    "ceiling": ceiling,
    "char": char,
    "div": div,
    "enconvert": enconvert,
    "equals": eql,
    'filenametotag': filenametotag,
    "find": find,
    "floor": floor,
    "format": formatValue,
    "geql": geql,
    "grtr": grtr,
    "hasformat": hasformat,
    "if": if_,
    "iflonger": iflonger,
    # 'image_to_file': image_to_file,
    'import_text': import_text,
    "isdigit": isdigit,
    "left": left,
    "len": len_,
    "leql": leql,
    "less": less,
    "lower": lower,
    'merge_values': merge_values,
    'meta_sep': meta_sep,
    'meta': meta,
    "mid": mid,
    "mod": mod,
    "move": move,
    "mul": mul,
    'remove_dupes': remove_dupes,
    "neql": neql,
    "not": not_,
    "num": num,
    "odd": odd,
    "or": or_,
    "rand": rand,
    "re_escape": re_escape,
    'remove_except': remove_except,
    # 'remove_tag': remove_tag,
    "replace": replace,
    "regex": replaceWithReg,
    "right": right,
    "round": _round,
    'save_artwork': save_artwork,
    'sort': sort_field,
    'split_by_sep': split_by_sep,
    "strip": strip,
    "sub": sub,
    'tag_dir': tag_dir,
    "texttotag": texttotag,
    'testfunction': testfunction,
    "titleCase": titleCase,
    'remove_fields': remove_fields,
    "upper": upper,
    'update_from_tag': update_from_tag,
    "validate": validate,
    'to_ascii': to_ascii,
    'to_num': to_num
}

no_fields = [filenametotag, load_images, move, remove_except,
             save_artwork, tag_dir, update_from_tag]
no_preview = [autonumbering, load_images, remove_tag, save_artwork]

from . import findfunc

FuncError = findfunc.FuncError
ParseError = findfunc.ParseError



import os

from PyQt5.QtCore import QMutex, Qt
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QGridLayout, QLabel, QScrollArea, QVBoxLayout, QWidget

from .. import audioinfo
from ..audioinfo import tag_versions
from ..constants import LEFTDOCK, SELECTIONCHANGED
from ..puddleobjects import PuddleThread, natural_sort_key

mutex = QMutex()


def sort_dict(d):
    ret = []
    for key, val in d.items():
        if isinstance(val, str):
            ret.append((key, val))
        else:
            ret.extend((key, v) for v in val)
    ret.sort(key=lambda v: natural_sort_key(v[0]))
    return ret


class StoredTags(QScrollArea):
    def __init__(self, parent=None, status=None):
        QScrollArea.__init__(self, parent)
        self.emits = []
        self.receives = self.receives = [(SELECTIONCHANGED, self.load)]
        self._labels = []
        font = QFont()
        font.setBold(True)
        self._boldfont = font
        self._init()
        self.setWidgetResizable(True)
        self._status = status
        self._loading = False
        self._lastfilepath = None

    def _init(self):
        widget = QWidget()
        self._grid = QGridLayout()
        self._grid.setColumnStretch(1, 1)
        self._grid.setContentsMargins(3, 3, 3, 3)
        widget.setLayout(self._grid)
        self.setWidget(widget)
        # widget.wheelEvent.connect(self._hScroll)

    def load(self):
        audios = self._status['selectedfiles']
        if not self.isVisible():
            return
        if audios:
            filepath = audios[0].filepath
            if self._status['previewmode'] and filepath == self._lastfilepath:
                return
            self._lastfilepath = filepath
        else:
            self._init()
            return

        def retrieve_tag():
            if not os.path.exists(filepath):
                return
            try:
                audio = audioinfo._Tag(filepath)
                tags = tag_versions.tags_in_file(filepath)
            except (OSError, IOError) as e:
                audio = {'Error': [e.strerror]}

            if isinstance(audio, audioinfo.id3.Tag):
                if 'ID3v2.4' in tags:
                    tags.remove('ID3v2.4')
                if 'ID3v2.3' in tags:
                    tags.remove('ID3v2.3')
                if 'ID3v2.2' in tags:
                    tags.remove('ID3v2.2')
            elif hasattr(audio, 'apev2') and audio.apev2:
                if 'APEv2' in tags:
                    tags.remove('APEv2')

            ret = [(audio['__tag_read'], sort_dict(audio.usertags))]

            for tag in tags:
                if not tag == audio['__tag_read']:
                    try:
                        ret.append((tag, sort_dict(tag_versions.tag_values(filepath, tag))))
                    except:
                        continue
            return ret

        def _load(tags):
            # print 'loading', time.time()
            while self._loading:
                QApplication.processEvents()
            try:
                self._loading = True
                self._init()
                if not tags:
                    return

                grid = self._grid

                offset = 1
                for title, values in tags:
                    if not values:
                        continue
                    t_label = QLabel(title)
                    t_label.setFont(self._boldfont)
                    grid.addWidget(t_label, offset - 1, 0)
                    for row, (tag, value) in enumerate(values):
                        field = QLabel('%s:' % tag)
                        field.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
                        grid.addWidget(field, row + offset, 0)
                        vlabel = QLabel(value)
                        grid.addWidget(vlabel, row + offset, 1)
                    grid.setRowMinimumHeight(grid.rowCount(),
                                             vlabel.sizeHint().height())
                    offset += grid.rowCount() + 1
                vbox = QVBoxLayout()
                vbox.addStretch()
                grid.addLayout(vbox, offset + 1 + offset, 0, -1, -1)
                grid.setRowStretch(offset + 1, 1)
            finally:
                self._loading = False

        thread = PuddleThread(retrieve_tag, self)
        thread.threadfinished.connect(_load)
        # print 'starting thread', time.time()
        thread.start()

    def showEvent(self, event):
        super(StoredTags, self).showEvent(event)
        self.load()

    def wheelEvent(self, e):
        h = self.horizontalScrollBar()
        if not self.verticalScrollBar().isVisible() and h.isVisible():
            numsteps = e.angleDelta().y() // 5
            h.setValue(h.value() - numsteps)
            e.accept()
        else:
            QScrollArea.wheelEvent(self, e)


control = ('Stored Tags', StoredTags, LEFTDOCK, False)



from PyQt5.QtWidgets import QApplication


def translate(context, source_text):
    return QApplication.translate(context, source_text)


def dont_execute():
    # General Translations
    translate("GenSettings", 'Su&bfolders')
    translate("GenSettings", 'Show &gridlines')
    translate("GenSettings", 'Show tooltips in file-view:')
    translate("GenSettings", 'Show &row numbers')
    translate("GenSettings", 'Automatically resize columns to contents')
    translate("GenSettings", '&Preserve file modification times')
    translate("GenSettings", 'Program to &play files with:')
    translate("GenSettings", '&Load last folder at startup')

    # Artwork
    translate("Artwork Context", 'Cover Varies')
    translate("Artwork Context", 'No Images')

    # Menus
    translate("Menus", 'Enabl&e Preview Mode')
    translate("Menus", 'Clear Selected &Files')
    translate("Menus", '&Write Previews')
    translate("Menus", '&Undo Last Clear')
    translate("Menus", 'Sort &By')
    translate("Menus", 'Clear Selected &Cells')
    translate("Menus", "&Plugins")
    translate("Menus", "Sort &By")

# Below here is generated by `update_translation.py`
    translate("Menus", "&Text File->Tag")
    translate("Menus", "&Format")
    translate("Menus", "Refresh &Selected")
    translate("Menus", "Reloads directories of selected files.")
    translate("Menus", "Remove ID3v&1 Tag")
    translate("Menus", "Remove ID3v&2 Tag")
    translate("Menus", "Select &Next Directory")
    translate("Menus", "&Lock Layout")
    translate("Menus", "&Unload Everything")
    translate("Menus", "Autonumbering &Wizard...")
    translate("Menus", "&Functions")
    translate("Menus", "&Preferences")
    translate("Menus", "&Actions")
    translate("Menus", "&Import Music Library...")
    translate("Menus", "&Paste")
    translate("Menus", "&Copy Selection")
    translate("Menus", "&Rename Directories")
    translate("Menus", "&QuickActions")
    translate("Menus", "&Properties")
    translate("Menus", "&Delete")
    translate("Menus", "Paste &Onto Selection")
    translate("Menus", "&Undo")
    translate("Menus", "&File->Tag")
    translate("Menus", "Convert filename to tag using the pattern.")
    translate("Menus", "Load &playlist")
    translate("Menus", "Import an m3u playlist into puddletag.")
    translate("Menus", "&Add Folder")
    translate("Menus", "Append a directory to current file-view.")
    translate("Menus", "&Open Folder")
    translate("Menus", "Select a directory to import into puddletag.")
    translate("Menus", "&Play")
    translate("Menus", "Plays the selected files in the predefined music player.")
    translate("Menus", "&Save")
    translate("Menus", "&Refresh")
    translate("Menus", "Refresh current file-view.")
    translate("Menus", "Sa&ve playlist")
    translate("Menus", "Save all files to m3u playlist.")
    translate("Menus", "E&xtended Tags")
    translate("Menus", "Move Selected Do&wn")
    translate("Menus", "Select &Previous Directory")
    translate("Menus", "Select all files belonging to the directories of those selected. Otherwise (if only a single directory is selected) selects all the files in the previous directory.")
    translate("Menus", "&Remove Tag")
    translate("Menus", "Copy All &Fields")
    translate("Menus", "In &Library")
    translate("Menus", "Replace...")
    translate("Menus", "Remove &APEv2 Tag")
    translate("Menus", "Remove All &ID3 Tags")
    translate("Menus", "Move Selected &Up")
    translate("Menus", "&Select Column")
    translate("Menus", "&Cut")
    translate("Menus", "&Increase Font")
    translate("Menus", "&Exit")
    translate("Menus", "&Tag->File")
    translate("Menus", "Select &All")
    translate("Menus", "&Invert Selection")
    translate("Menus", "&Decrease Font")
    translate("Menus", "&Clipboard->Tag")
    translate("Menus", "&Go to parent folder")
    translate("Menus", "Focuses the current file's parent folder in the Filesystem view")
    translate("Menus", "&File")
    translate("Menus", "&Edit")
    translate("Menus", "&Convert")
    translate("Menus", "&Actions")
    translate("Menus", "&Tools")
    translate("Menus", "&Preview Mode")
    translate("Menus", "Ta&g Tools")
    translate("Menus", "&Windows")

    # Functions
    translate("Functions", "Artwork: Filenames='$1', Description='$2', Case Sensitive=$3")
    translate("Functions", "Load Artwork")
    translate("Functions", "&Filenames to check (;-separated, shell wildcards [eg. *] allowed)")
    translate("Functions", "&Default description (can be pattern):")
    translate("Functions", "Match filename's &case:")
    translate("Functions", "Autonumbering: $0, Start: $1, Restart for dir: $2, Padding: $3")
    translate("Functions", "Autonumbering")
    translate("Functions", "oi")
    translate("Functions", "1")
    translate("Functions", "aoeu")
    translate("Functions", "False")
    translate("Functions", "au")
    translate("Functions", "1")
    translate("Functions", "Convert to encoding: $0, Encoding: $1")
    translate("Functions", "Convert from non-standard encoding")
    translate("Functions", "&Encoding")
    translate("Functions", "cp1250")
    translate("Functions", "cp1251")
    translate("Functions", "cp1252")
    translate("Functions", "cp1253")
    translate("Functions", "cp1254")
    translate("Functions", "cp1255")
    translate("Functions", "cp1256")
    translate("Functions", "cp1257")
    translate("Functions", "cp1258")
    translate("Functions", "euc_jp")
    translate("Functions", "cp932")
    translate("Functions", "euc_jis_2004")
    translate("Functions", "shift_jis")
    translate("Functions", "johab")
    translate("Functions", "big5")
    translate("Functions", "big5hkscs")
    translate("Functions", "gb2312")
    translate("Functions", "gb18030")
    translate("Functions", "gbk")
    translate("Functions", "hz")
    translate("Functions", "File->Tag '$1'")
    translate("Functions", "Filename to Tag")
    translate("Functions", "&Pattern")
    translate("Functions", "Format $0 using $1")
    translate("Functions", "Format value")
    translate("Functions", "&Format string")
    translate("Functions", "Text File: $0, '$1'")
    translate("Functions", "Import text file")
    translate("Functions", "&Pattern (can be relative path)")
    translate("Functions", "lyrics.txt")
    translate("Functions", "Merge field: $0, sep='$1'")
    translate("Functions", "Merge field")
    translate("Functions", "&Separator")
    translate("Functions", ";")
    translate("Functions", "Tag->File: $1")
    translate("Functions", "Tag to filename")
    translate("Functions", "&Pattern")
    translate("Functions", "Remove Dupes: $0, Match Case $1")
    translate("Functions", "Remove duplicate values")
    translate("Functions", "Match &Case")
    translate("Functions", "Remove fields except: $1")
    translate("Functions", "Remove all fields except")
    translate("Functions", "&Field list (; separated):")
    translate("Functions", "")
    translate("Functions", "Replace $0: '$1' -> '$2', Match Case: $3, Words Only: $4")
    translate("Functions", "Replace")
    translate("Functions", "&Replace")
    translate("Functions", "w&ith:")
    translate("Functions", "Match c&ase:")
    translate("Functions", "only as &whole word")
    translate("Functions", "RegReplace $0: RegExp '$1' with '$2', Match Case: $3")
    translate("Functions", "Replace with RegExp")
    translate("Functions", "&Regular Expression")
    translate("Functions", "Replace &matches with:")
    translate("Functions", "Match &Case")
    translate("Functions", "Export Art: pattern='$1'")
    translate("Functions", "Export artwork to file")
    translate("Functions", "&Pattern (extension not required)")
    translate("Functions", "folder_%img_counter%")
    translate("Functions", "Sort $0, order='$1', Match Case='$2'")
    translate("Functions", "Sort values")
    translate("Functions", "&Order")
    translate("Functions", "Ascending")
    translate("Functions", "Descending")
    translate("Functions", "Match &Case")
    translate("Functions", "Split using separator $0: sep='$1'")
    translate("Functions", "Split fields using separator")
    translate("Functions", "&Separator")
    translate("Functions", ";")
    translate("Functions", "Trim $0")
    translate("Functions", "Trim whitespace")
    translate("Functions", "Tag->Dir: $1")
    translate("Functions", "Tag to Dir")
    translate("Functions", "&Pattern (can be relative path)")
    translate("Functions", "%artist% - %album%")
    translate("Functions", "Text to Tag: $0 -> $1, $2")
    translate("Functions", "Text to Tag")
    translate("Functions", "&Text")
    translate("Functions", "&Pattern")
    translate("Functions", "&Output")
    translate("Functions", "Convert Case: $0: $1")
    translate("Functions", "Case conversion")
    translate("Functions", "&Type")
    translate("Functions", "Mixed Case")
    translate("Functions", "UPPER CASE")
    translate("Functions", "lower case")
    translate("Functions", "For &Mixed Case, after any of:")
    translate("Functions", "., !")
    translate("Functions", "<blank> $0")
    translate("Functions", "Remove Fields")
    translate("Functions", "Update from $2, Fields: $1")
    translate("Functions", "Update from tag")
    translate("Functions", "&Field list (; separated):")
    translate("Functions", "&Tag")
    translate("Functions", "APEv2")
    translate("Functions", "ID3")
    translate("Functions", "Converts all unicode chars to ASCII.")

    # Dialogs
    translate("Dialogs", "Tag Panel")
    translate("Dialogs", "Artwork")
    translate("Dialogs", "Filesystem")
    translate("Dialogs", "Filter")
    translate("Dialogs", "Tag Sources")
    translate("Dialogs", "Stored Tags")
    translate("Dialogs", "Logs")
    translate("Dialogs", "Mass Tagging")
    translate("Dialogs", "Functions")
    translate("Dialogs", "Actions")



# -*- coding: utf-8 -*-
# rhythmbox.py

# Copyright (C) 2008-2009 concentricpuddle

# This file is part of puddletag, a semi-good music tag editor.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import urllib.error
import urllib.parse
import urllib.request
from collections import defaultdict
from os import path
from xml.sax import make_parser
from xml.sax.handler import ContentHandler

from PyQt5.QtCore import QDir, QSettings, QUrl
from PyQt5.QtWidgets import QFileDialog, QHBoxLayout, QLabel, QLineEdit, QPushButton, QVBoxLayout, QWidget

from .. import audioinfo
from .. import musiclib
from ..util import translate

FILENAME, PATH = audioinfo.FILENAME, audioinfo.PATH

name = "Rhythmbox"
description = "Rhythmbox Database"
author = 'concentricpuddle'


def getFilename(filename):
    filename = urllib.request.url2pathname(filename)

    if filename.startswith('file://'):
        filename = filename[len('file://'):]
        return {'__dirpath': path.dirname(filename),
                PATH: filename,
                FILENAME: path.basename(filename),
                "__ext": path.splitext(filename)[1][1:],
                "__dirname": path.basename(path.dirname(filename))}


getTime = lambda date: audioinfo.strtime(int(date))
getCreated = lambda created: {'__created': getTime(created)}
getModified = lambda modified: {'__modified': getTime(modified)}
getLength = lambda length: {'__length': audioinfo.strlength(int(length))}
getBitRate = lambda bitrate: {'__bitrate': bitrate + ' kb/s'}

CONVERSION = {'title': 'title',
              'genre': 'genre',
              'artist': 'artist',
              'album': 'album',
              'track-number': 'track',
              'duration': getLength,
              'file-size': '__size',
              'location': getFilename,
              'first-seen': getCreated,
              'mtime': getModified,
              'last-seen': '__last_seen',
              'bitrate': getBitRate,
              'disc-number': 'discnumber'}

setLength = lambda length: {'duration': str(audioinfo.lnglength(length))}
setCreated = lambda created: {'first-seen': str(audioinfo.lngtime(created))}
setBitrate = lambda bitrate: {'bitrate': str(audioinfo.lngfrequency(bitrate) / 1000)}
setModified = lambda modified: {'last-seen': str(audioinfo.lngtime(modified))}
setFilename = lambda filename: {'location': 'file://' + str(QUrl.toPercentEncoding(filename, '/()"\'')).encode('utf8')}

RECONVERSION = {
    'title': 'title',
    'artist': 'artist',
    'album': 'album',
    'track': 'track-number',
    'discnumber': 'disc-number',
    'genre': 'genre',
    '__length': setLength,
    '__created': setCreated,
    '__bitrate': setBitrate,
    '__modified': setModified,
    '__filename': setFilename,
    '__size': 'file-size'}

SUPPORTEDTAGS = ['artist', 'genre', 'title', 'track', '__size', 'album']


class DBParser(ContentHandler):
    indent = " " * 4

    def __init__(self):
        # Information is stored as follows:
        # self.albums is a dictionary with each key being an artist.
        # self.albums[key] is also a dictionary with album names as keys
        # and an integer specifying the index of the album in self.tracks
        # self.tracks being a list of lists, each of which contains the
        # track metadata for an album as dictionaries.
        self.tagval = ""
        self.name = ""
        self.stargetting = False
        self.values = {}
        self.current = "#nothing"
        self.tracks = []
        self.albums = defaultdict(lambda: {})
        self.extravalues = []
        self.extras = False
        self.extratype = ""

    def characters(self, ch):
        try:
            self.values[self.current] += ch
        except KeyError:
            self.values[self.current] = ch

    def endElement(self, name):
        if name == "entry" and self.stargetting:
            self.stargetting = False

            if not self.extras:
                tag = {}
                for field, value in self.values.items():
                    try:
                        tag.update(CONVERSION[field](value.strip()))
                    except TypeError:
                        tag[CONVERSION[field]] = value.strip()
                    except KeyError:
                        tag['#' + field] = value.strip()

                f = ((k, v.strip()) for k, v in tag.items())
                tag = dict((k, v) for k, v in f if v)

                album = tag.get('album', '')
                artist = tag.get('artist', '')

                albums = self.albums[artist]
                if album not in albums:
                    albums[album] = len(self.tracks)
                    self.tracks.append([tag])
                else:
                    self.tracks[albums[album]].append(tag)
            else:
                x = ((k, v.strip()) for k, v in self.values.items())
                x = dict((k, v) for k, v in x if v)
                x['name'] = self.extratype
                self.extratype = ""
                self.extravalues.append(x)
                self.extras = False
            self.values = {}

    def _escapedText(self, txt):
        result = txt
        result = result.replace("&", "&amp;")
        result = result.replace("<", "&lt;")
        result = result.replace(">", "&gt;")
        return result

    def parse_file(self, filename):
        parser = make_parser()
        parser.setContentHandler(self)
        try:
            parser.parse(filename)
        except ValueError as detail:
            if not os.path.exists(filename):
                msg = "%s does not exist." % filename
            else:
                msg = "%s is not a valid Rhythmbox XML database." % filename
            raise musiclib.MusicLibError(0, msg)
        except (IOError, OSError) as detail:
            if not os.path.exists(filename):
                msg = "%s does not exist." % filename
            else:
                msg = detail.strerror()
            raise musiclib.MusicLibError(0, msg)
        self.filename = filename
        return self.albums, self.tracks

    def startElement(self, name, attrs):
        def startelement(name, attrs):
            if name == 'entry':
                if attrs.get('type') == 'song':
                    self.stargetting = True
                else:
                    self.extratype = attrs.get('type')
                    self.extras = True
                    self.stargetting = True
            if self.stargetting and name != 'entry':
                self.current = name
                self.values[name] = ""

        if name == 'rhythmdb':
            version = attrs.get('version')
            self.head = '<?xml version="1.0" standalone="yes"?>\n' \
                        '  <rhythmdb version="%s">' % str(version)
            self.startElement = startelement


class RhythmDB(ContentHandler):
    indent = " " * 4

    def __init__(self, filename):
        # Information is stored as follows:
        # self.albums is a dictionary with each key being an artist.
        # self.albums[key] is also a dictionary with album names as keys
        # and an integer specifying the index of the album in self.tracks
        # self.tracks being a list of lists, each of which contains the
        # track metadata for an album as dictionaries.
        self.tagval = ""
        self.name = ""
        self.stargetting = False
        self.values = {}
        self.current = "nothing"
        self.tracks = []
        self.albums = {}
        self.extravalues = []
        self.extras = False
        self.extratype = ""
        parser = make_parser()
        parser.setContentHandler(self)
        try:
            parser.parse(filename)
        except ValueError as detail:
            if not os.path.exists(filename):
                msg = "%s does not exist." % filename
            else:
                msg = "%s is not a valid Rhythmbox XML database." % filename
            raise musiclib.MusicLibError(0, msg)
        except (IOError, OSError) as detail:
            if not os.path.exists(filename):
                msg = "%s does not exist." % filename
            else:
                msg = detail.strerror()
            raise musiclib.MusicLibError(0, msg)
        self.filename = filename

    def startElement(self, name, attrs):
        def startelement(name, attrs):
            if name == 'entry':
                if attrs.get('type') == 'song':
                    self.stargetting = True
                else:
                    self.extratype = attrs.get('type')
                    self.extras = True
                    self.stargetting = True
            if self.stargetting and name != 'entry':
                self.current = name
                self.values[name] = ""

        if name == 'rhythmdb':
            version = attrs.get('version')
            self.head = '<?xml version="1.0" standalone="yes"?>\n' \
                        '  <rhythmdb version="%s">' % str(version)
            self.startElement = startelement

    def characters(self, ch):
        try:
            self.values[self.current] += ch
        except KeyError:
            self.values[self.current] = ch

    def endElement(self, name):
        if name == "entry" and self.stargetting:
            self.stargetting = False
            if not self.extras:
                audio = {}
                for tag, value in self.values.items():
                    try:
                        audio.update(CONVERSION[tag](value.strip()))
                    except TypeError:
                        audio[CONVERSION[tag]] = value.strip()
                    except KeyError:
                        audio["___" + tag] = value.strip()
                audio['__library'] = 'rhythmbox'

                if audio['artist'] not in self.albums:
                    self.albums[audio['artist']] = {}
                albums = self.albums[audio['artist']]
                if audio['album'] not in albums:
                    albums[audio['album']] = len(self.tracks)
                    self.tracks.append([audio])
                else:
                    index = albums[audio['album']]
                    self.tracks[index].append(audio)
            else:
                x = dict([(z, v.strip()) for z, v in self.values.items()])
                x['name'] = self.extratype
                self.extratype = ""
                self.extravalues.append(x)
                self.extras = False
            self.values = {}

    def tracksByTag(self, parent, parentvalue, child=None, childval=None):
        if parent not in SUPPORTEDTAGS:
            return
        if parent == 'artist' and child == 'album':
            return self.getTracks(parentvalue, childval)

        if (childval is None) or (child is None):
            files = []
            for album in self.tracks:
                for f in album:
                    if f[parent] == parentvalue:
                        files.append(f)
        elif childval and child:
            files = []
            for album in self.tracks:
                for f in album:
                    if f[parent] == parentvalue and f[child] == childval:
                        files.append(f)

        return [musiclib.Tag(self, z) for z in files]

    def children(self, parent, parentvalue, child):
        if parent == 'artist' and child == 'album':
            return self.getAlbums(parentvalue)
        else:
            values = set()
            for album in self.tracks:
                [values.add(z[child]) for z in album if z[parent] == parentvalue]
            return list(values)

    def distinctValues(self, tag):
        if tag not in SUPPORTEDTAGS:
            return
        if tag == 'artist':
            return list(self.albums.keys())
        else:
            values = set()
            for album in self.tracks:
                [values.add(z[tag]) for z in album]
            return list(values)

    def getArtists(self):
        return list(self.albums.keys())

    def getAlbums(self, artist):
        try:
            return list(self.albums[artist].keys())
        except KeyError:
            return None

    def getTracks(self, artist, albums=None):
        ret = []
        if albums is None:
            albums = list(self.albums[artist].keys())

        if artist in self.albums:
            stored = self.albums[artist]
            for album in albums:
                if album in stored:
                    ret.extend(self.tracks[stored[album]])
        return [musiclib.Tag(self, z) for z in ret]

    def _escapedText(self, txt):
        result = txt
        result = result.replace("&", "&amp;")
        result = result.replace("<", "&lt;")
        result = result.replace(">", "&gt;")
        return result

    def delTracks(self, tracks):
        prevartist = None
        prevalbum = None
        for track in tracks:
            track = audioinfo.stringtags(track)
            artist = track['artist']
            album = track['album']
            if artist != prevartist or album != prevalbum:
                dbtracks = self.tracks[self.albums[artist][album]]
                filenames = [z[FILENAME] for z in dbtracks]
            del (dbtracks[filenames.index(track[FILENAME])])
            filenames.remove(track[FILENAME])
            if not dbtracks:
                del (self.albums[artist][album])
            if not self.albums[artist]:
                del (self.albums[artist])

    def saveTracks(self, tracks):
        for old, new in tracks:
            old, new = audioinfo.stringtags(old), audioinfo.stringtags(new)
            artist = new['artist']
            album = new['album']
            if old['artist'] != artist:
                if artist in self.albums:
                    if album in self.albums[artist]:
                        index = self.albums[artist][album]
                        self.tracks[index].append(new)
                    else:
                        self.albums[artist][album] = len(self.tracks)
                        self.tracks.append([new])
                else:
                    self.albums[artist] = {album: len(self.tracks)}
                    self.tracks.append([new])
            elif album != old['album']:
                if album in self.albums[artist]:
                    self.albums[artist][album].append(new)
                else:
                    self.albums[artist][album] = len(self.tracks)
                    self.tracks.append([new])
            else:
                self.tracks[self.albums[artist][album]].append(new)
            self.delTracks([old])

    def save(self):
        filename = path.join(path.dirname(self.filename), 'rhythmbox.xml')
        f = open(filename, 'w')
        entry = [self.head + "\n"]
        for album in self.tracks:
            for track in album:
                entry.append('  <entry type="song">\n')
                for key, tagvalue in track.items():
                    try:
                        if key.startswith('___'):
                            tagname = key[len('___'):]
                        else:
                            temp = RECONVERSION[key](tagvalue)
                            tagname = list(temp.keys())[0]
                            tagvalue = temp[tagname]
                    except TypeError:
                        tagname = RECONVERSION[key]
                    except KeyError:
                        continue
                    entry.append('    <%s>%s</%s>\n' % (self._escapedText(tagname), self._escapedText(tagvalue), self._escapedText(tagname)))
                entry.append('  </entry>\n')
                f.write(("".join(entry)))
                entry = []

        entry = []
        for value in self.extravalues:
            entry.append('  <entry type ="%s">\n' % value['name'])
            [entry.append('    <%s>%s</%s>\n' %
                          (self._escapedText(val), self._escapedText(value[val]),
                           self._escapedText(val))) for val in value]
            entry.append("  </entry>\n")
            f.write(("".join(entry)))
            entry = []
        f.write("</rhythmdb>")
        f.close()
        backup = path.join(path.dirname(self.filename), 'oldrhythmdb.xml')
        if not path.exists(backup):
            os.rename(self.filename, backup)
        os.rename(filename, self.filename)

    def close(self):
        pass

    def search(self, term):
        term = term.upper()
        ret = []
        artists = set([z for z in self.albums if term in z.upper()])
        [ret.extend(self.getTracks(z)) for z in artists]
        others = set(self.albums).difference(artists)

        for artist in others:
            albums = self.albums[artist]
            for album in albums:
                if term in album.upper():
                    ret.extend(self.getTracks(artist, [album]))
                else:
                    index = self.albums[artist][album]
                    tracks = self.tracks[index]
                    for track in tracks:
                        for value in track.values():
                            if term in value.upper():
                                ret.append(track)
                                break
        return [musiclib.Tag(self, z) for z in ret]

    def updateSearch(self, term, tracks):
        tags = ['artist', 'title', FILENAME, '__path', 'album', 'genre',
                'comment', 'year']
        term = term.lower()
        tracks = []
        for audio in files:
            temp = audioinfo.stringtags(audio)
            for tag in tags:
                if term in temp[tag].lower():
                    tracks.append(audio)
                    break
        return tracks


class InitWidget(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.dbpath = QLineEdit(path.join(str(QDir.homePath()), ".gnome2/rhythmbox/rhythmdb.xml"))

        vbox = QVBoxLayout()
        label = QLabel('&Database Path')
        label.setBuddy(self.dbpath)
        [vbox.addWidget(z) for z in [label, self.dbpath]]

        hbox = QHBoxLayout()
        openfile = QPushButton("&Browse...")
        hbox.addStretch()
        hbox.addWidget(openfile)
        vbox.addLayout(hbox)
        openfile.clicked.connect(self.getFile)
        vbox.addStretch()
        self.setLayout(vbox)
        self.dbpath.selectAll()
        self.dbpath.setFocus()

    def getFile(self):
        selectedFile = QFileDialog.getOpenFileName(self,
                                                   'Select RhythmBox database file.', self.dbpath.text())
        filename = selectedFile[0]
        if filename:
            self.dbpath.setText(filename)

    def library(self):
        dbpath = str(self.dbpath.text())
        try:
            return RhythmDB(dbpath)
        except musiclib.MusicLibError as e:
            raise e
        except Exception as e:
            raise musiclib.MusicLibError(0,
                translate("Rhythmbox", '{} is an invalid Rhythmbox music library file.').format(dbpath)
                ) from e


if __name__ == '__main__':
    k = DBParser()
    x, y = k.parse_file('rdb.xml')
    # import pdb
    # pdb.set_trace()
    print([i for i, z in enumerate(y) if len(z) > 1])
    print(list(x.keys())[0], x[list(x.keys())[0]], y[34])



import struct

import mutagen.id3
from mutagen.id3 import ParseID3v1

from . import apev2
from . import id3

APEv2_Tag = apev2.Tag

_v2_nums = set([2, 3, 4])

ID3_V1 = 'id3_v1'
ID3_V2 = 'id3_v2'
APEv2 = 'ape_v2'

TAG_TYPES = [ID3_V1, ID3_V2, APEv2]


def apev2_values(fn):
    assert isinstance(fn, str)

    return APEv2_Tag(fn).usertags


def convert_id3_frames(frames):
    mapping = id3.Tag.mapping
    return dict((mapping.get(k, k), v.get_value())
                for k, v in id3.handle(frames)).items()


def fullread(fileobj, size):
    data = fileobj.read(size)
    if len(data) != size: raise EOFError
    return data


def has_apev2(fn):
    fileobj = open(fn, 'rb') if isinstance(fn, str) else fn

    try:
        fileobj.seek(-160, 2)
    except IOError:
        return False

    footer = fileobj.read()
    return b"APETAGEX" in footer


def has_v1(fn):
    close_file = isinstance(fn, str)
    fileobj = open(fn, 'rb') if close_file else fn

    try:
        fileobj.seek(-128, 2)
        return "TAG" == struct.unpack("3s", fullread(fileobj, 3))[0]
    except (struct.error, EOFError, EnvironmentError):
        return False
    finally:
        if close_file:
            fileobj.close()


def get_v2(fn):
    close_file = isinstance(fn, str)
    fileobj = open(fn, 'rb') if close_file else fn

    size = 5
    try:
        id3, vmaj, vrev = struct.unpack('>3sBB', fullread(fileobj, size))
    except EOFError:
        return
    finally:
        if close_file:
            fileobj.close()

    if id3 == 'ID3' and vmaj in _v2_nums:
        return (2, vmaj, vrev) if vrev != 0 else (2, vmaj)
    return


def id3v1_values(fn):
    close_file = isinstance(fn, str)
    fileobj = open(fn, 'rb') if close_file else fn

    fileobj.seek(-128, 2)
    frames = ParseID3v1(fileobj.read(128))
    if close_file:
        fileobj.close()
    if frames:
        return convert_id3_frames(frames)


def id3v2_values(fn):
    assert isinstance(fn, str)

    try:
        frames = mutagen.id3.ID3(fn)
    except:
        return None
    if frames:
        return convert_id3_frames(frames)


def id3_tags(fn):
    close_file = isinstance(fn, str)
    fileobj = open(fn, 'rb') if close_file else fn
    version = []

    try:
        version = [(1, 1)] if has_v1(fileobj) else []
    except EOFError:
        if close_file:
            fileobj.close()
        return []

    fileobj.seek(0)
    v2 = get_v2(fileobj)
    if v2:
        version.append(v2)
    if close_file:
        fileobj.close()
    return version


def tags_in_file(fn, to_check=(ID3_V1, ID3_V2, APEv2)):
    fileobj = open(fn, 'rb') if isinstance(fn, str) else fn

    if ID3_V1 in to_check and ID3_V2 in to_check:
        tags = ['ID3v' + '.'.join(map(str, z)) for z in id3_tags(fileobj)]
    elif ID3_V1 in to_check:
        tags = ['ID3v1.1'] if has_v1(fileobj) else []
    elif ID3_V2 in to_check:
        tags = get_v2(fileobj)
        tags = ['ID3v' + '.'.join(map(str, tags))] if tags else []
    else:
        tags = []

    if APEv2 in to_check and has_apev2(fn):
        tags.append('APEv2')
    return tags


_value_types = {
    APEv2: apev2_values,
    ID3_V1: id3v1_values,
    ID3_V2: id3v2_values, }


def tag_values(fn, tag):
    tag = tag.lower()
    if tag.startswith('id3v1'):
        tag = ID3_V1
    elif tag.startswith('id3v2'):
        tag = ID3_V2
    elif tag.startswith('ape'):
        tag = APEv2

    if tag not in TAG_TYPES:
        return {}

    return _value_types[tag](fn)


if __name__ == '__main__':
    import sys

    filename = sys.argv[1]
    # f = open(filename, 'rb')
    print(tags_in_file(filename))
    print(id3v1_values(filename))
    print(id3v2_values(filename))
    print(apev2_values(filename))



from copy import deepcopy

import re

from html.parser import HTMLParser

from puddlestuff.functions import replace_regex
from puddlestuff.audioinfo import CaselessDict

conditionals = set(['if', 'ifnot'])


def debug(cursor, flag, filename=None, maxsize=None):
    if flag == 'on':
        cursor.debug = True
        cursor.debug_file = filename
    else:
        cursor.debug = False
    return


def do(cursor):
    cursor._domodified = deepcopy(cursor.output)
    cursor.num_loop += 1
    return


def _else(cursor):
    num_if = 0

    for i, (command, l, a) in enumerate(cursor.source[cursor.cmd_index + 1:]):
        if command == 'if' or command == 'ifnot':
            num_if += 1
        elif command == 'endif':
            if num_if == 0:
                cursor.next_cmd = i + cursor.cmd_index + 2
                return True
            else:
                num_if -= 1


def endif(cursor): return


def findinline(cursor, text, n=1, exit=None):
    cursor.log('findinline %s %d\n' % (text, n))
    line = cursor.line[cursor.charno:]
    i = 1
    t_len = len(text)
    start = 0
    while line.find(text, start) != -1:
        pos = line.find(text, start) + len(text)
        start = pos
        if i == n:
            cursor.charno += pos
            return
        i += 1
    cursor.charno = len(cursor.line) - 1
    return


def findline(cursor, text, index=1, exit=None, no_case=False):
    num_found = 1
    if no_case:
        text = text.lower()
    if index < 0:
        index = 0 - index
        if no_case:
            lines = reversed(cursor.all_lowered[:cursor.lineno])
        else:
            lines = reversed(cursor.all_lines[:cursor.lineno])
        for i, line in enumerate(lines):
            if text in line:
                if num_found == index:
                    cursor.log('Found %s at line %d\n' % (text, i))
                    cursor.lineno = cursor.lineno - i - 1
                    cursor.charno = 0
                    return
                else:
                    num_found += 1
    else:
        if no_case:
            lines = cursor.lowered
        else:
            lines = cursor.lines
        for i, line in enumerate(lines):
            if text in line:
                if num_found == index:
                    cursor.log('Found %s at line %d\n' % (text, i))
                    cursor.lineno = cursor.lineno + i
                    cursor.charno = 0
                    return
                else:
                    num_found += 1

    cursor.lineno = len(cursor.all_lines) - 1


def findlinenocase(cursor, text, num=1, exit=None):
    return findline(cursor, text, num, text, True)


def gotochar(cursor, num):
    cursor.charno = num - 1


def gotoline(cursor, num):
    cursor.lineno = num - 1
    cursor.charno = num - 1


def _if(cursor, text, ifnot=False):
    if ifnot:
        if not cursor.line[cursor.charno:].startswith(text):
            return
    else:
        if cursor.line[cursor.charno:].startswith(text):
            return

    num_if = 0
    for i, (command, lineno, args) in enumerate(cursor.source[cursor.cmd_index + 1:]):
        if command in ('if', 'ifnot'):
            num_if += 1
            # print 'if', cursor.source[i + cursor.cmd_index + 1]
        elif command == 'endif':
            # print 'end_if', cursor.source[i + cursor.cmd_index + 1]
            if num_if == 0:
                cursor.next_cmd = i + cursor.cmd_index + 2
                # print 'endif', cursor.source[i + cursor.cmd_index + 1]
                return True
            else:
                num_if -= 1
        elif command == 'else' and num_if == 0:
            # print 'else', cursor.source[i + cursor.cmd_index + 1]
            cursor.next_cmd = i + cursor.cmd_index + 2
            # print 'else', cursor.source[i + cursor.cmd_index + 2]
            # pdb.set_trace()
            return True


def ifnot(cursor, text):
    return _if(cursor, text, True)


def joinlines(cursor, num):
    ret = cursor.lines[:num]
    cursor.all_lines[cursor.lineno] = ''.join(ret)
    del (cursor.all_lines[cursor.lineno: cursor.lineno + num])
    cursor.lineno = cursor.lineno


def joinuntil(cursor, text):
    ret = []
    index = None
    for line in cursor.lines:
        if text in line:
            index = line.find(text) + len(text)
            ret.append(line[:index])
            # if line.strip() == text:
            break
        else:
            v = line.strip()
            if v:
                ret.append(v)

    if index is None:
        return

    append = [line[index:]] if line[index:].strip() else []
    al = cursor.all_lines
    cursor.all_lines = al[:cursor.lineno] + [''.join(ret)] + \
                       append + al[cursor.lineno + len(ret):]
    cursor.lineno = cursor.lineno


def killtag(cursor, tag, repl=' '):
    if repl:
        cursor.log('Killing HTML tag %s with %s\n' % (tag, repl))
    else:
        cursor.log('Killing HTML tag %s\n' % tag)

    if tag == '*':
        parser = TagProcessor()
        parser.feed(cursor.line)
        text = '%s%s%s' % (' ', repl.join(parser.pieces), ' ')
        cursor.log(cursor.line + ' becomes ' + text + '\n')
        cursor.line = text
        return
    else:
        leave, to_rep = cursor.line[:cursor.charno], cursor.line[cursor.charno:]
        cursor.line = leave + to_rep.replace('<%s>' % tag, repl)


def movechar(cursor, num):
    cursor.charno = cursor.charno + num


def moveline(cursor, num, exit=None):
    cursor.log('Moving to line %d' % (cursor.lineno + num))
    cursor.lineno = cursor.lineno + num
    cursor.charno = 0


def outputto(cursor, text):
    if text.lower() == 'tracks' and cursor.num_loop:
        if cursor.output is not cursor.album:
            cursor.tracks.append(cursor.output)
        cursor.output = CaselessDict()
        cursor.field = 'title'
    elif text.lower() == 'tracks':
        if cursor.field in cursor.output:
            cursor.output[cursor.field] += cursor.cache
        else:
            cursor.output[cursor.field] = cursor.cache
        field = cursor.field
        cursor.cache = ''
        cursor.tracks = {}
        cursor.output = CaselessDict()
        cursor.field = 'title'
        del (cursor.output[field])
    elif not cursor.num_loop:
        if cursor.output is not cursor.album:
            cursor.output = cursor.album
        cursor.log('Outputting %s' % text)
        cursor.field = text
    else:
        cursor.log('Outputting %s' % text)
        cursor.field = text


def replace(cursor, s, repl):
    text = cursor.line[cursor.charno:].replace(s, repl)
    cursor.line = cursor.line[:cursor.charno] + text


def regexpreplace(cursor, regexp, s):
    text = replace_regex({}, cursor.line, regexp, s,
                         matchcase=True)
    # Now uses whole line instead of just from the current char onwards
    # because Mp3tag is being a fucking idiot.
    cursor.line = text
    cursor.charno = 0


def say(cursor, text):
    cursor.log('say %s' % text)
    cursor.cache += text


def saynchars(cursor, num):
    cursor.cache += cursor.line[cursor.charno + 1: cursor.charno + num]
    cursor.charno += num


def saynewline(cursor):
    cursor.cache += '\n'


def saynextnumber(cursor):
    try:
        number = re.search(r'\d+', cursor.line[cursor.charno:]).group()
        cursor.log('Saying number %s\n' % number)
        cursor.cache += number
        cursor.charno += len(number)
    except AttributeError:
        return


def saynextword(cursor):
    word = re.search(r'\w+', cursor.line[cursor.charno:]).group()
    cursor.cache += word
    cursor.charno += len(word)


def sayoutput(cursor, field):
    if field.lower().strip() == 'tracks' and cursor.tracks:
        field = 'title'

    track_field = [field.lower(), cursor.field.lower()]
    track_field = cursor.track_fields.intersection(track_field)
    track_field = track_field or (cursor.field.lower() == 'track')

    if track_field and cursor.tracks:
        field = field.lower()
        if field in cursor.output:
            v = [z.strip() for z in cursor.output[field].split("|")]
            v_len = len(v)
        for i, t in enumerate(cursor.tracks):
            if field in t:
                t[cursor.field] = t.get(field)
            elif v and i < v_len:
                t[cursor.field] = v[i]
    else:
        v = cursor.output.get(field, '')
        cursor.cache += v


def sayregexp(cursor, rexp, separator=None, check=None):
    if (check is not None) and (check not in cursor.line):
        return
    if cursor.charno + 1 == len(cursor.line):
        return
    if check:
        line = cursor.line[cursor.charno: cursor.line.find(check, cursor.charno)]
    else:
        line = cursor.line[cursor.charno:]
    if separator is not None:
        matches = [match.group() for match in
                   re.finditer(rexp, line)]
        cursor.cache += separator.join(matches)
    else:
        match = re.search(rexp, line).group()
        if match:
            cursor.cache += match


def sayrest(cursor):
    cursor.log('Saying the rest of line from position %d.' % cursor.charno)
    cursor.log('Line: %s.' % cursor.line)
    cursor.log('Saying: %s\n' % cursor.line[cursor.charno:])
    cursor.cache += cursor.line[cursor.charno:]
    cursor.charno = len(cursor.line) - 1


def sayuntil(cursor, text):
    cursor.log('SayUntil start: %d' % cursor.charno)
    cursor.log('Line: %s.' % cursor.line)
    line = cursor.line[cursor.charno:]
    index = line.find(text)
    if index != -1:
        cursor.cache += line[:index]
        cursor.log('Saying: %s\n' % line[:index])
        cursor.charno = cursor.charno + index
    else:
        cursor.log('%s not found.\n' % text)
        cursor.log('Saying: %s\n' % line)
        cursor.cache += line
        cursor.charno = len(cursor.line) - 1


def sayuntilml(cursor, text):
    cache = []
    lines = cursor.lines[::]
    lines[0] = cursor.line[cursor.charno:]
    for i, line in enumerate(lines):
        index = line.find(text)
        if index == -1:
            cache.append(line)
        else:
            cache.append(line[:index])
            cursor.cache += '\n'.join(cache)
            cursor.charno = index + 1
            cursor.lineno += i
            return


def _set(cursor, field, value=None):
    if not value:
        if field in cursor.output:
            del (cursor.output[field])
        elif cursor.field == field:
            cursor.cache = ""
    else:
        cursor.output[field] = value if \
            isinstance(value, str) else str(value)


def _while(cursor, condition, numtimes=None):
    cursor.num_loop -= 1
    cursor.num_iters = 0
    nested = 0
    if cursor.line[cursor.charno:].startswith(condition):
        for i, (cmd, lineno, args) in enumerate(reversed(cursor.source[:cursor.cmd_index])):
            if cmd == 'do':
                if nested > 0:
                    nested -= 1
                    continue
                if numtimes is None:
                    cursor.next_cmd = cursor.cmd_index - i - 1
                    return True
                elif numtimes >= cursor.num_iters:
                    cursor.next_cmd = cursor.cmd_index - i - 1
                    return True
                cursor.num_iters += 1
            elif cmd == 'while':
                nested += 1

    if cursor.tracks == {}:
        cursor.tracks = [{'title': z.strip()} for z in
                         cursor.cache.split('|')]
        cursor.track_fields.add('title')
    elif cursor.output and cursor.tracks:
        if cursor.output != cursor._domodified:
            cursor.tracks.append(cursor.output)
            list(map(cursor.track_fields.add, (z.lower() for z in cursor.output)))


def unspace(cursor):
    cursor.line = cursor.line.strip()
    cursor.charno = 0


class TagProcessor(HTMLParser):

    def reset(self):
        self.pieces = []
        HTMLParser.reset(self)

    def handle_data(self, text):
        self.pieces.append(text)


FUNCTIONS = {
    'debug': debug,
    'do': do,
    'else': _else,
    'endif': endif,
    'findinline': findinline,
    'findline': findline,
    'findlinenocase': findlinenocase,
    'gotochar': gotochar,
    'gotoline': gotoline,
    'if': _if,
    'ifnot': ifnot,
    'joinlines': joinlines,
    'joinuntil': joinuntil,
    'killtag': killtag,
    'movechar': movechar,
    'moveline': moveline,
    'outputto': outputto,
    'regexpreplace': regexpreplace,
    'replace': replace,
    'say': say,
    'saynchars': saynchars,
    'saynewline': saynewline,
    'saynextnumber': saynextnumber,
    'saynextword': saynextword,
    'sayoutput': sayoutput,
    'sayregexp': sayregexp,
    'sayrest': sayrest,
    'sayuntil': sayuntil,
    'sayuntilml': sayuntilml,
    'set': _set,
    'unspace': unspace,
    'while': _while}



from functools import partial

from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtWidgets import QAction, QMenu

from ..puddleobjects import PuddleConfig

from copy import deepcopy
from ..constants import FILESSELECTED
from ..plugins import connect_shortcut
from ..translations import translate

status = {}

_previews = []
_sort_action = None

ENABLED = translate("Menus", 'Enabl&e Preview Mode')
DISABLED = translate("Menus", 'Disabl&e Preview Mode')


class PreviewAction(QAction):
    def setEnabled(self, value):
        if status['previewmode'] and value:
            super(PreviewAction, self).setEnabled(True)
        else:
            super(PreviewAction, self).setEnabled(False)


def toggle_preview_display(action, preview_actions, value):
    if value:
        action.setText(DISABLED)
        files = status['selectedfiles']
        for p in preview_actions:
            if files:
                p.setEnabled(True)
    else:
        action.setText(ENABLED)
        global _previews
        _previews = []
        for p in preview_actions:
            p.setEnabled(False)


def toggle_preview_mode():
    action = QObject().sender()
    if action.text() == ENABLED:
        emit('enable_preview_mode')
    else:
        emit('disable_preview_mode')


def clear_selected():
    files = status['selectedfiles']
    _previews.append(dict([(f, deepcopy(f.preview)) for f in files]))
    emit('setpreview', [{} for f in files])


def clear_selected_cells():
    files = status['selectedfiles']
    selected = status['selectedtags']

    _previews.append(dict([(f, f.preview) for f in files]))

    ret = []
    for fields, f in zip(selected, files):
        ret.append(dict([(k, v) for k, v in f.preview.items()
                         if k not in fields]))
    emit('setpreview', ret)


def create_actions(parent):
    enable_preview = QAction('Enabl&e Preview Mode', parent)
    enable_preview.setShortcut('Ctrl+Shift+P')
    enable_preview.triggered.connect(toggle_preview_mode)

    clear_selection = PreviewAction('Clear Selected &Files', parent)
    clear_selection.setShortcut('Ctrl+Shift+F')
    clear_selection.triggered.connect(clear_selected)

    write = PreviewAction('&Write Previews', parent)
    write.setShortcut('Ctrl+W')

    write.triggered.connect(lambda: emit('writepreview'))

    revert = PreviewAction('&Undo Last Clear', parent)
    revert.setShortcut('Ctrl+Shift+Z')
    revert.triggered.connect(undo_last)

    sort = QAction('Sort &By', parent)
    sort.triggered.connect(sort_by_fields)

    clear_cells = PreviewAction('Clear Selected &Cells', parent)
    clear_cells.triggered.connect(clear_selected_cells)

    cparser = PuddleConfig()
    options = cparser.get('table', 'sortoptions',
                          ['__filename,track,__dirpath', 'track, album',
                           '__filename,album,__dirpath'])
    global _sort_action
    _sort_action = sort
    sort_actions = set_sort_options(options)

    preview_actions = [clear_selection, write, revert, clear_cells]

    toggle = partial(toggle_preview_display, enable_preview, preview_actions)

    obj.receives.append(['previewModeChanged', toggle])

    [connect_shortcut(z, FILESSELECTED) for z in preview_actions]

    return [enable_preview, clear_selection, write, revert, sort,
            clear_cells] + sort_actions


def set_sort_options(options):
    parent = _sort_action.parentWidget()
    menu = QMenu(parent)
    sort_actions = []
    options = [[z.strip() for z in option.split(',')] for option in options]
    for option in options:
        action = QAction('/'.join(option), parent)
        action.sortOption = option
        menu.addAction(action)
        action.triggered.connect(sort_by_fields)
        sort_actions.append(action)
    _sort_action.setMenu(menu)
    status['sort_actions'] = sort_actions
    return sort_actions


def sort_by_fields():
    options = QObject().sender().sortOption
    files = status['selectedfiles']
    model = status['table'].model()
    if files and len(files) > 1:
        model.sortByFields(options,
                           status['selectedfiles'], status['selectedrows'])
    else:
        model.sortByFields(options)


def set_status(stat):
    global status
    status = stat


def undo_last():
    if _previews:
        emit('setpreview', _previews.pop())


class _SignalObject(QObject):
    enable_preview_mode = pyqtSignal(name='enable_preview_mode')
    disable_preview_mode = pyqtSignal(name='disable_preview_mode')
    setpreview = pyqtSignal(object, name='setpreview')
    writepreview = pyqtSignal(name='writepreview')


obj = _SignalObject()
obj.emits = ['enable_preview_mode', 'disable_preview_mode', 'setpreview',
             'writepreview']
obj.receives = []


def emit(sig, *args):
    getattr(obj, sig).emit(*args)



import logging
import logging.config

from .constants import LOG_FILENAME


def init_logger(level):
    config = {
        "version": 1,
        "formatters": {
            "simple": {
                "format": '[%(asctime)s]%(levelname)s:%(message)s',
            }
        },
        "handlers": {
            "console": {
                "class": 'logging.StreamHandler',
                "level": level if level == logging.DEBUG else logging.ERROR,
                "formatter": 'simple',
                "stream": 'ext://sys.stdout',
            },
            "file": {
                "class": 'logging.handlers.RotatingFileHandler',
                "level": level,
                "formatter": 'simple',
                'filename': LOG_FILENAME,
                'maxBytes': 1024 * 1024 * 10  # 10MB
            }
        },
        "root": {
            "level": level,
            "handlers": ['file', 'console'],
        }
    }
    logging.config.dictConfig(config)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2005 Michael Urman
# Copyright (C) 2006-2008, 2011-2012 Lukáš Lalinský
# Copyright (C) 2013-2014 Sophist-UK
# Copyright (C) 2013-2014, 2018 Laurent Monin
# Copyright (C) 2014, 2018-2020 Philipp Wolfer
# Copyright (C) 2016 Christoph Reiter
# Copyright (C) 2016 Ville Skyttä
# Copyright (C) 2017 Sambhav Kothari
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from mutagen.id3 import (
    ID3,
    Frames,
    Frames_2_2,
    TextFrame,
)


try:
    from mutagen.id3 import GRP1
except ImportError:
    class GRP1(TextFrame):
        pass


class XSOP(TextFrame):
    pass


known_frames = dict(Frames)
known_frames.update(dict(Frames_2_2))
known_frames["GRP1"] = GRP1  # Available since mutagen >= 1.38
known_frames["XSOP"] = XSOP


class CompatID3(ID3):

    """
    Additional features over mutagen.id3.ID3:
     * Allow some v2.4 frames also in v2.3
    """

    PEDANTIC = False

    def __init__(self, *args, **kwargs):
        if args:
            kwargs["known_frames"] = known_frames
        super().__init__(*args, **kwargs)

    def update_to_v23(self):
        update_to_v23(self)


def update_to_v23(tags):
    # leave TSOP, TSOA and TSOT even though they are officially defined
    # only in ID3v2.4, because most applications use them also in ID3v2.3
    frames = []
    for key in ["TSOP", "TSOA", "TSOT", "TSST"]:
        frames.extend(tags.getall(key))
    ID3.update_to_v23(tags)
    for frame in frames:
        tags.add(frame)



import logging
from collections import defaultdict

status = {}


def connect_shortcut(action, enabled, disabled=None, togglecheck=None):
    controls = status['dialogs']
    emits = defaultdict(lambda: [])

    for c in controls.values():
        [emits[sig].append(c) for sig in c.emits]

    if enabled in emits:
        [getattr(c, enabled).connect(
            action.setEnabled) for c in emits[enabled]]
    else:
        logging.error('No enable signal found for ' + action.text())
        action.setEnabled(False)

    if togglecheck and togglecheck in emits:
        [getattr(c, togglecheck).connect(
            action.setEnabled) for c in emits[togglecheck]]


def connect_control(control):
    controls = status['dialogs']
    emits = defaultdict(lambda: [])

    for c in controls.values():
        [emits[sig].append(c) for sig in c.emits]

    for signal, slot in control.receives:
        if signal in emits:
            [getattr(c, signal).connect(slot) for c in emits[signal]]

    for c in controls.values():
        for signal, slot in c.receives:
            if signal in control.emits:
                getattr(control, signal).connect(slot)



from collections import defaultdict
from copy import deepcopy
from functools import partial

import mutagen.id3 as id3
import mutagen.mp3

try:
    from mutagen.aiff import AIFF, _IFFID3
except ImportError:
    AIFF = None

try:
    from mutagen.dsf import DSF, _DSFID3
except ImportError:
    DSF = None

from mutagen.id3 import (APIC, PairedTextFrame, TextFrame, TimeStampTextFrame,
                         UrlFrame)

from . import util
from ._compatid3 import CompatID3

from .constants import MODES
from .util import (CaselessDict, FILENAME, MockTag, PATH,
                   cover_info, del_deco, fn_hash, get_mime, get_total,
                   getdeco, info_to_dict, isempty, keys_deco, parse_image, set_total,
                   setdeco, str_filesize, unicode_list, usertags)

TagBase = MockTag


# Placed here to prohibit the circular import from tag_versions.
def handle(audio):
    """Converts mutagen.id3.ID3 object to format puddletag understands.

    Returns a dictionary with the following a puddletag fields
    as keys eg. 'title'. Values associated with each field contain
    the mutagen.id3.Frame associated with that field.

    Each Frame object is modified to have get_value and set_value
    methods. Some frames will also have a 'frames' attributes.
    """

    # See the comment in for 'Tag.__tags' in 'Tag.__init__' for more info.
    handlers = defaultdict(lambda: [])

    for frame in audio.values():
        if frame.__class__ in frames:
            handlers[frames[frame.__class__]].append(frame)

    keys = {}
    ret = {}

    # Handler returns a dictionary with fields as used in puddletag
    # eg 'title' and text, a frame, or a list of frames as its
    # associated value.
    for handler, frame in handlers.items():
        for k, v in handler(frame).items():
            lower = k.lower()
            if lower in keys:
                ret[keys[lower]].append(v) if isinstance(v, str) \
                    else ret[keys[lower]].extend(v)
            else:
                keys[lower] = k
                ret[k] = v
    return ret


from . import tag_versions

ATTRIBUTES = ('frequency', 'length', 'bitrate', 'accessed', 'size', 'created',
              'modified', 'filetype')

WRITE_V1 = 1
WRITE_V2 = 2
WRITE_BOTH = 3

v1_option = 2
v2_option = 4
apev2_option = False

ISO8859 = 0
UTF16 = 1
UTF16BE = 2
UTF8 = 3

encoding = UTF8


def get_factory(func, frame):
    return lambda: func(frame)


def set_factory(func, frame):
    return lambda value: func(frame, value)


def create_text(title, value):
    frame = revtext_frames[title](encoding, value)
    frame.get_value = lambda: get_text(frame)
    frame.set_value = partial(set_text, frame)
    return {title: frame}


def get_text(textframe):
    return [str(z) for z in textframe.text]


def set_text(frame, value):
    frame.text = TextFrame(encoding, value).text
    frame.encoding = encoding
    return True


def text_handler(title):
    def func(frames):
        frame = frames[0]
        frame.get_value = lambda: get_text(frame)
        frame.set_value = partial(set_text, frame)
        return {title: frame}

    return func


def create_genre(value):
    frame = id3.TCON(encoding, value)
    frame.get_value = lambda: get_genre(frame)
    frame.set_value = lambda value: set_genre(frame, value)
    return {'genre': frame}


def get_genre(frame):
    return frame.genres


def set_genre(frame, value):
    frame.genres = value
    frame.encoding = encoding
    return True


def genre_handler(frames):
    frame = frames[0]
    frame.get_value = lambda: get_genre(frame)
    frame.set_value = lambda value: set_genre(frame, value)
    return {'genre': frame}


text_frames = {
    id3.TALB: 'album',
    id3.TBPM: 'bpm',
    id3.TCOM: 'composer',
    id3.TCOP: "copyright",
    id3.TDAT: "date",
    id3.TDLY: "audiodelay",
    id3.TENC: "encodedby",
    id3.TEXT: "lyricist",
    id3.TFLT: "filetype",
    id3.TIME: "time",
    id3.TIT1: "grouping",
    id3.TIT2: "title",
    id3.TIT3: "version",
    id3.TKEY: "initialkey",
    id3.TLAN: "language",
    id3.TLEN: "audiolength",
    id3.TMED: "mediatype",
    id3.TMOO: "mood",
    id3.TOAL: "originalalbum",
    id3.TOFN: "filename",
    id3.TOLY: "author",
    id3.TOPE: "originalartist",
    id3.TORY: "originalyear",
    id3.TOWN: "fileowner",
    id3.TPE1: "artist",
    id3.TPE2: "albumartist",
    id3.TPE3: "conductor",
    id3.TPE4: "arranger",
    id3.TPOS: "discnumber",
    id3.TPRO: "producednotice",
    id3.TPUB: "organization",
    id3.TRCK: "track",
    id3.TRDA: "recordingdates",
    id3.TRSN: "radiostationname",
    id3.TRSO: "radioowner",
    id3.TSIZ: "audiosize",
    id3.TSO2: "albumartistsortorder",
    id3.TSOA: "albumsortorder",
    id3.TSOP: "performersortorder",
    id3.TSOT: "titlesortorder",
    id3.TSRC: "isrc",
    id3.TSSE: "encodingsettings",
    id3.TSST: "setsubtitle",
    id3.TYER: 'year',
}

try:
    text_frames.update({
        id3.TCMP: "itunescompilationflag",
        id3.TSOC: "itunescomposersortorder"})
except AttributeError:
    pass

revtext_frames = dict((key, frame) for frame, key in text_frames.items())
write_frames = dict((key, partial(create_text, key)) for
                    key in text_frames.values())
write_frames['genre'] = create_genre


def create_time(title, value):
    frame = revtime_frames[title](encoding)
    if not set_time(frame, value):
        return {}
    frame.get_value = lambda: get_text(frame)
    frame.set_value = partial(set_time, frame)
    return {title: frame}


def set_time(frame, value):
    text = TimeStampTextFrame(encoding, value).text
    if not [_f for _f in text if _f]:
        return
    frame.text = text
    frame.encoding = encoding
    return True


def time_handler(title):
    def func(frames):
        frame = frames[0]
        frame.get_value = lambda: get_text(frame)
        frame.set_value = partial(set_time, frame)
        return {title: frame}

    return func


time_frames = {
    id3.TDEN: "encodingtime",
    id3.TDOR: "originalreleasetime",
    id3.TDRC: "year",
    id3.TDRL: "releasetime",
    id3.TDTG: "taggingtime"}

revtime_frames = dict([(key, frame) for frame, key in time_frames.items()])
write_frames.update([(key, partial(create_time, key)) for
                     key in revtime_frames])


def create_usertext(title, value):
    frame = id3.TXXX(encoding, title, value)
    frame.get_value = lambda: get_text(frame)
    frame.set_value = partial(set_text, frame)
    return {title: frame}


def usertext_handler(frames):
    d = {}
    for frame in frames:
        frame.get_value = get_factory(get_text, frame)
        frame.set_value = set_factory(set_text, frame)
        d[frame.desc] = frame
    return d


url_frames = {
    id3.WCOP: "wwwcopyright",
    id3.WOAF: "wwwfileinfo",
    id3.WOAS: "wwwsource",
    id3.WORS: "wwwradio",
    id3.WPAY: "wwwpayment",
    id3.WPUB: "wwwpublisher"}


def create_url(title, value):
    frame = revurl_frames[title]()
    frame.get_value = lambda: get_url(frame)
    frame.set_value = partial(set_url, frame)
    frame.set_value(value)
    return {title: frame}


def get_url(frame):
    return [frame.url]


def set_url(frame, value):
    if not isinstance(value, str):
        value = value[0]
    frame.url = UrlFrame(value).url
    return True


def url_handler(title):
    def func(frames):
        frame = frames[0]
        frame.get_value = lambda: get_url(frame)
        frame.set_value = partial(set_url, frame)
        return {title: frame}

    return func


revurl_frames = dict([(key, frame) for frame, key in url_frames.items()])
write_frames.update([(key, partial(create_url, key)) for key in revurl_frames])

uurl_frames = {
    id3.WCOM: "wwwcommercialinfo",
    id3.WOAR: "wwwartist"}


def create_uurl(title, value):
    frame = revuurl_frames[title]()
    d = uurl_handler(title)([frame])
    d[title].set_value(value)
    return d


def uurl_handler(title):
    def set_uurl(frames, value):
        if isinstance(value, str):
            value = [value]
        while frames:
            frames.pop()
        cls = revuurl_frames[title]
        frames.extend([cls(v) for v in value])

    def func(frames):
        frame = frames[0]
        frame.get_value = lambda: [f.url for f in frames]
        frame.set_value = partial(set_uurl, frames)
        frame.frames = frames
        return {title: frame}

    return func


revuurl_frames = dict([(key, frame) for frame, key in uurl_frames.items()])
write_frames.update([(key, partial(create_uurl, key)) for
                     key in revuurl_frames])


def create_userurl(title, value):
    value = to_string(value)
    desc = title[len('www:'):]
    frame = id3.WXXX(encoding, desc, value)
    frame.get_value = lambda: get_url(frame)
    frame.set_value = partial(set_url, frame)
    return {title: frame}


def userurl_handler(frames):
    d = {}
    for frame in frames:
        frame.get_value = get_factory(get_url, frame)
        frame.set_value = set_factory(set_url, frame)
        d['www:' + frame.desc] = frame
    return d


paired_textframes = {
    id3.TIPL: "involvedpeople",
    id3.TMCL: "musiciancredits",
    id3.IPLS: "involvedpeople"}


def create_paired(key, value):
    frame = revpaired_frames[key](encoding)
    if set_paired(frame, value):
        frame.get_value = get_factory(get_paired, frame)
        frame.set_value = set_factory(set_paired, frame)
        return {key: frame}
    return {}


def get_paired(frame):
    return [';'.join([':'.join(z) for z in frame.people])]


def set_paired(frame, text):
    if not isinstance(text, str):
        text = text[0]
    value = [people.split(':') for people in text.split(';')]
    temp = []
    for pair in value:
        if len(pair) == 1:
            temp.append([pair[0], ''])
        else:
            temp.append(pair)
    value = temp
    frame.people = PairedTextFrame(encoding, value).people
    return True


def paired_handler(title):
    def func(frame):
        frame = frame[0]
        frame.get_value = lambda: get_paired(frame)
        frame.set_value = partial(set_paired, frame)
        return {title: frame}

    return func


revpaired_frames = dict([(key, frame) for frame,
                                          key in paired_textframes.items()])
write_frames.update([(key, partial(create_paired, key)) for
                     key in revpaired_frames])


def create_comment(desc, value):
    frame = id3.COMM(encoding, 'XXX', desc, value)
    frame.get_value = lambda: get_text(frame)
    frame.set_value = partial(set_text, frame)
    return {'comment:' + frame.desc: frame}


def set_commentattrs(frame):
    frame.get_value = lambda: get_text(frame)
    frame.set_value = partial(set_text, frame)


def comment_handler(frames):
    d = {}
    for frame in frames:
        set_commentattrs(frame)
        if not frame.desc and 'comment' not in d:
            d['comment'] = frame
        else:
            d['comment:' + frame.desc] = frame
    return d


def create_playcount(value):
    frame = id3.PCNT()
    if set_playcount(frame, value):
        frame.get_value = lambda: get_playcount(frame)
        frame.set_value = partial(set_playcount, frame)
        return {'playcount': frame}
    return {}


def get_playcount(frame):
    return [str(frame.count)]


def set_playcount(frame, value):
    if not isinstance(value, str):
        value = value[0]
    try:
        frame.count = int(value)
    except ValueError:
        return
    return True


def playcount_handler(frame):
    frame = frame[0]
    frame.get_value = lambda: get_playcount(frame)
    frame.set_value = partial(set_playcount, frame)
    return {'playcount': frame}


def create_popm(values):
    if isinstance(values, str):
        values = [values]
    frames = [_f for _f in [set_popm(id3.POPM(), v) for v in values] if _f]
    if frames:
        return popm_handler(frames)
    return {}


def get_popm(frame):
    if not hasattr(frame, 'count'):
        frame.count = 0
    return ':'.join([frame.email, str(frame.rating),
                     str(frame.count)])


def to_string(value):
    if isinstance(value, bytes):
        return value.decode('utf8')
    elif isinstance(value, str):
        return value
    else:
        return to_string(value[0])


def set_popm(frame, value):
    value = to_string(value)
    try:
        email, rating, count = value.split(':', encoding)
        rating = int(rating)
        count = int(count)
    except ValueError:
        return
    frame.email = email
    frame.rating = rating
    frame.count = count
    return frame


def popm_handler(frames):
    def set_values(frames, values):
        if isinstance(values, str):
            values = [values]
        temp = [_f for _f in [set_popm(id3.POPM(), v) for v in values] if _f]
        if not temp:
            return
        while frames:
            frames.pop()
        [frames.append(v) for v in temp]

    get_value = lambda: [get_popm(f) for f in frames]
    set_value = partial(set_values, frames)

    for frame in frames:
        frame.get_value = get_value
        frame.set_value = set_value
        frame.frames = frames
    return {'popularimeter': frame}


def create_ufid(key, value):
    if not isinstance(value, bytes):
        try:
            value = value[0]
        except IndexError:
            return {}
    if isinstance(value, str):
        value = value.encode('utf8')
    owner = key[len('ufid:'):]
    frame = id3.UFID(owner, value)
    frame.get_value = partial(get_ufid, frame)
    frame.set_value = partial(set_ufid, frame)
    return {'ufid:' + frame.owner: frame}


def set_ufid(frame, value):
    if not isinstance(value, bytes):
        try:
            value = value[0]
        except IndexError:
            return {}
    if isinstance(value, str):
        value = value.encode('utf8')
    frame.data = value


def get_ufid(frame):
    return [frame.data.decode('utf8')]


def ufid_handler(frames):
    d = {}
    for frame in frames:
        try:
            frame.data.decode('utf8')
        except UnicodeDecodeError:
            continue
        frame.get_value = get_factory(get_ufid, frame)
        frame.set_value = set_factory(set_ufid, frame)
        d['ufid:' + frame.owner] = frame
    return d


def _parse_rgain(value):
    if not isinstance(value, str):
        try:
            value = value[0]
        except IndexError:
            return

    values = [z.strip() for z in value.split(':')]
    channel, gain, peak = values
    channel = int(channel)
    gain = float(gain)
    peak = float(peak)
    return channel, gain, peak


def create_rgain(key, value):
    desc = key[len('rgain:'):]
    try:
        channel, gain, peak = _parse_rgain(value)
    except (TypeError, ValueError):
        return {}

    frame = id3.RVA2(desc, channel, gain, peak)
    frame.get_value = get_factory(get_rgain, frame)
    frame.set_value = set_factory(set_rgain, frame)

    return {'rgain:' + desc: frame}


def set_rgain(frame, value):
    try:
        channel, gain, peak = _parse_rgain(value)
    except (TypeError, ValueError):
        return {}
    frame.channel = channel
    frame.gain = gain
    frame.peak = peak


def get_rgain(frame):
    return ':'.join(map(str, [frame.channel, frame.gain, frame.peak]))


def rgain_handler(frames):
    d = {}
    for f in frames:
        f.get_value = get_factory(get_rgain, f)
        f.set_value = set_factory(set_rgain, f)
        d['rgain:' + f.desc] = f
    return d


def create_uslt(value):
    f = id3.USLT()
    set_uslt(f, value)
    f.set_value = set_factory(set_uslt, f)
    if f.frames:
        return {'unsyncedlyrics': f}
    return {}


def set_uslt(f, value):
    if isinstance(value, str):
        value = [value]

    frames = []

    for lyrics in value:
        try:
            lyrics = [z for z in lyrics.split('|', 3)]
        except (TypeError, ValueError):
            continue
        len_l = len(lyrics)
        if len_l == 1:
            lang = ''
            desc = ''
            text = lyrics[0]
        elif len_l == 2:
            lang = lyrics[0].strip()
            desc = ''
            text = lyrics[1]
        elif len_l == 3:
            lang, desc, text = lyrics
            lang = lang.strip()
        elif len_l > 3:
            text = ''.join(lyrics[2:])
            lang = lyrics[0].strip()
            desc = lyrics[1]
        else:
            continue

        if not lang or len(lang) != 3:
            lang = 'und'
        frames.append(id3.USLT(encoding, lang, desc, text))

    if not frames:
        f.frames = []
        f.get_value
        return {}

    f.frames = frames
    f.get_value = get_uslt(frames)


def get_uslt(frames):
    def text(f, attr):
        ret = getattr(f, attr, '')
        return ret if isinstance(ret, str) else \
            str(ret, 'utf8', 'replace')

    ret = ['%s|%s|%s' % (text(frame, 'lang'), text(frame, 'desc'),
                          text(frame, 'text')) for frame in frames]
    return lambda: ret


def uslt_handler(frames):
    d = {}
    f = frames[0]
    f.get_value = get_uslt(frames)
    f.set_value = set_factory(set_uslt, f)
    f.frames = frames
    d['unsyncedlyrics'] = f
    return d


write_frames.update({
    'playcount': create_playcount,
    'popularimeter': create_popm,
    'genre': create_genre,
    'unsyncedlyrics': create_uslt})

# Dictionary contaning classes from mutagen.id3 and it's associated
# handler.
frames = dict([(key, text_handler(title)) for key,
                                              title in text_frames.items()])

frames.update([(key, time_handler(title)) for key, title in
               time_frames.items()])

frames.update([(key, url_handler(title)) for key, title in
               url_frames.items()])

frames.update([(key, uurl_handler(title)) for key, title in
               uurl_frames.items()])

frames.update([(key, paired_handler(title)) for key, title in
               paired_textframes.items()])

frames.update({
    id3.TCON: genre_handler,
    id3.WXXX: userurl_handler,
    id3.TXXX: usertext_handler,
    id3.COMM: comment_handler,
    id3.PCNT: playcount_handler,
    id3.POPM: popm_handler,
    id3.UFID: ufid_handler,
    id3.RVA2: rgain_handler,
    id3.USLT: uslt_handler, })

revframes = dict((v, k) for k, v in frames.items())


def bin_to_pic(image):
    return {'data': image.data, 'description': image.desc,
            'mime': image.mime, 'imagetype': image.type}


def pic_to_bin(image):
    data = image[util.DATA]
    description = image.get(util.DESCRIPTION)
    if not description:
        description = ''
    imagetype = image.get(util.IMAGETYPE, encoding)
    mime = image.get(util.MIMETYPE)
    if not mime:
        mime = get_mime(data)
    return APIC(encoding, mime, imagetype, description, data)


class ID3(CompatID3):
    """ID3 reader to replace mutagen's just to allow the reading of APIC
    tags with the same description, ala Mp3tag."""
    PEDANTIC = True

    def loaded_frame(self, tag):
        if len(type(tag).__name__) == encoding:
            tag = type(tag).__base__(tag)
        i = 0
        try:
            desc = tag.desc
            while tag.HashKey in self:
                tag.desc = desc + str(i)
                i += 1
        except AttributeError:
            "Nothing to do"
        self[tag.HashKey] = tag


class ID3FileType(mutagen.mp3.MP3):
    """See ID3 class."""

    def add_tags(self, ID3=ID3):
        return super(ID3FileType, self).add_tags(ID3)

    def load(self, filename, ID3=ID3, **kwargs):
        return super(ID3FileType, self).load(filename, ID3, **kwargs)


if AIFF is not None:
    class AIFFFileType(AIFF):
        """See ID3 class."""

        def add_tags(self, ID3=_IFFID3):
            AIFF.add_tags(self)

        def load(self, filename, **kwargs):
            return AIFF.load(self, filename, **kwargs)

if DSF is not None:
    class DSFFileType(DSF):
        """See ID3 class."""

        def add_tags(self, ID3=_IFFID3):
            DSF.add_tags(self)

        def load(self, filename, **kwargs):
            return DSF.load(self, filename, **kwargs)


def tag_factory(id3_filetype):
    class Tag(TagBase):
        IMAGETAGS = (util.MIMETYPE, util.DESCRIPTION, util.DATA,
                     util.IMAGETYPE)
        mapping = {}
        revmapping = {}

        def __init__(self, filename=None):
            self.__images = []

            self.__tags = CaselessDict()  # Used as storage.
            # Each key as the is the field as used by puddletag, eg. 'artist'
            # Each value contains an mutagen.id3.Frame object
            # that have two methods, get_value and set_value.
            # get_value returns the value stored by the frame as
            # text/unicode list.
            # set_value should take text/unicode list in parse it into
            # it understands.

            # When saving the frame stored will be used. If it has a 'frames'
            # attributes, those frames will be used instead.

            util.MockTag.__init__(self, filename)

        @property
        def filepath(self):
            return util.MockTag.get_filepath(self)

        @filepath.setter
        def filepath(self, val):
            self.__tags.update(util.MockTag.set_filepath(self, val))

        @property
        def images(self):
            return self.__images

        @images.setter
        def images(self, images):
            if images:
                self.__images = [parse_image(i, self.IMAGETAGS) for i in images]
            else:
                self.__images = []
            cover_info(images, self.__tags)

        @property
        def info(self):
            info = self.mut_obj.info
            fileinfo = [('Path', self[PATH]),
                        ('Size', str_filesize(int(self.size))),
                        ('Filename', self[FILENAME]),
                        ('Modified', self.modified)]
            try:
                version = self.mut_obj.tags.version
                version = ('ID3 Version', self.filetype)
            except AttributeError:
                version = ('ID3 Version', 'No tags in file.')
            fileinfo.append(version)

            if isinstance(self.mut_obj, DSFFileType):
                mpginfo = [('Type', 'DSF')]
            elif isinstance(self.mut_obj, AIFFFileType):
                mpginfo = [('Type', 'AIFF')]
            elif isinstance(self.mut_obj, ID3FileType):
                mpginfo = [('Version', 'MPEG %i Layer %i' % (info.version, info.layer))]
            else:
                mpginfo = []

            try:
                mpginfo.append(('Bitrate', self.bitrate))
            except AttributeError:
                pass

            try:
                mpginfo.append(('Frequency', self.frequency))
            except AttributeError:
                pass

            try:
                mpginfo.append(('Mode', MODES[info.mode]))
            except AttributeError:
                pass

            try:
                mpginfo.append(('Length', self.length))
            except Att:
                pass

            return [('File', fileinfo), (mpginfo[0][0], mpginfo)]

        def __contains__(self, key):
            if key == '__image':
                return bool(self.images)

            elif key == '__total':
                try:
                    return bool(get_total(self))
                except (KeyError, ValueError):
                    return False

            if self.revmapping:
                key = self.revmapping.get(key, key)

            return key in self.__tags

        def __deepcopy__(self, memo=None):
            cls = Tag()
            tags = CaselessDict()
            frames = []
            [frames.append(frame) if not hasattr(frame, 'frames') else
             frames.extend(frame.frames) for key, frame in self.__tags.items()
             if not key.startswith('__')]
            funcs = []
            frames_copy = []
            for frame in frames:
                funcs.append((getattr(frame, 'get_value', None),
                              getattr(frame, 'set_value', None)))
                if hasattr(frame, 'get_value'):
                    delattr(frame, 'get_value')
                if hasattr(frame, 'set_value'):
                    delattr(frame, 'set_value')
                frames_copy.append(deepcopy(frame))
            tags = handle(dict([(frame.HashKey, frame) for frame in frames_copy]))
            for frame, (get_value, set_value) in zip(frames, funcs):
                if get_value is not None:
                    frame.get_value = get_value
                if set_value is not None:
                    frame.set_value = set_value
            for key, value in self.__tags.items():
                if key not in tags:
                    tags[key] = deepcopy(value)
            cls.set_fundamentals(tags,
                                 self.mut_obj, deepcopy(self.images))
            cls.filepath = self.filepath
            return cls

        @del_deco
        def __delitem__(self, key):
            if key == '__image':
                self.images = []
            elif key.startswith('__'):
                return
            else:
                del (self.__tags[key])

        @getdeco
        def __getitem__(self, key):
            if key.startswith('__'):
                if key == '__image':
                    return self.images
                elif key == '__filetype':
                    return self.filetype
                elif key == '__total':
                    return get_total(self)
                else:
                    return self.__tags[key]
            elif not isinstance(key, str):
                return self.__tags[key]
            else:
                return self.__tags[key].get_value()

        @setdeco
        def __setitem__(self, key, value):
            if not isinstance(key, str):
                self.__tags[key] = value
                return
            if key.startswith('__'):
                if key == '__image':
                    self.images = value
                elif key in fn_hash:
                    setattr(self, fn_hash[key], value)
                elif key == '__total':
                    set_total(self, value)
                return

            value = unicode_list(value)

            if isempty(value):
                if key in self:
                    del (self[key])
                return

            if key in self.__tags:
                self.__tags[key].set_value(value)
            else:
                if key in write_frames:
                    self.__tags.update(write_frames[key](value))
                elif key == 'comment':
                    frame = {'comment': create_comment('', value)['comment:']}
                    self.__tags.update(frame)
                elif key.startswith('comment:'):
                    self.__tags.update(create_comment(key[len('comment:'):], value))
                elif key.startswith('www:'):
                    self.__tags.update(create_userurl(key, value))
                elif key.startswith('ufid:'):
                    self.__tags.update(create_ufid(key, value))
                elif key.startswith('rgain:'):
                    self.__tags.update(create_rgain(key, value))
                elif key.startswith('unsyncedlyrics:'):
                    self.__tags.update(create_uslt(key, value))
                else:
                    self.__tags.update(create_usertext(key, value))

        def delete(self):
            self.mut_obj.delete()
            for key in self.usertags:
                del (self.__tags[self.revmapping.get(key, key)])
            self.images = []

        @keys_deco
        def keys(self):
            return list(self.__tags.keys())

        def link(self, filename):
            """Links the audio, filename
            returns self if successful, None otherwise."""
            self.__images = []
            tags, audio = self.load(filename, id3_filetype)
            if audio is None:
                return

            if audio.tags:  # Not empty
                audio.tags.update_to_v24()
                self.__tags.update(handle(audio))

                # Get the image data.
                apics = audio.tags.getall("APIC")
                if apics:
                    self.images = list(map(bin_to_pic, apics))
                else:
                    self.images = []

            self.__tags.update(tags)
            self.__tags.update(info_to_dict(audio.info))

            if self.ext.lower() == 'mp3':
                self.__tags['__filetype'] = 'MP3'
            else:
                self.__tags['__filetype'] = 'ID3'

            self.set_attrs(ATTRIBUTES, self.__tags)

            try:
                self.__tags['__tag_read'] = 'ID3v%s.%s' % audio.tags.version[:2]
            except AttributeError:
                self.__tags['__tag_read'] = ''
            self.mut_obj = audio
            self._originaltags = list(audio.keys())
            self.update_tag_list()
            return self

        def save(self, v1=None, v2=None):
            if v1 is None:
                v1 = v1_option
            """Writes the tags to file."""
            filename = self.filepath
            if self.mut_obj.tags is None:
                self.mut_obj.add_tags()
            if filename != self.mut_obj.filename:
                self.mut_obj.tags.filename = filename
                self.mut_obj.filename = filename
            audio = self.mut_obj
            util.MockTag.save(self)

            userkeys = list(usertags(self.__tags).keys())
            frames = []
            [frames.append(frame) if not hasattr(frame, 'frames') else
             frames.extend(frame.frames) for key, frame in self.__tags.items()
             if key in userkeys]
            hashes = dict([(frame.HashKey, frame) for frame in frames])
            toremove = [z for z in self._originaltags if z in audio
                        and not (z in hashes or z.startswith('APIC'))]
            audio.update(hashes)

            old_apics = [z for z in audio if z.startswith('APIC')]
            if self.__images:
                newimages = []
                for image in [_f for _f in map(pic_to_bin, self.__images) if _f]:
                    i = 0
                    while image.HashKey in newimages:
                        i += 1
                        # Pad with spaces so that each key is unique.
                        image.desc += ' ' * i
                    audio[image.HashKey] = image
                    newimages.append(image.HashKey)
                [toremove.append(z) for z in old_apics if z not in newimages]
            else:
                toremove.extend(old_apics)

            for z in set(toremove):
                try:
                    del (audio[z])
                except KeyError:
                    continue

            audio.tags.filename = self.filepath
            v1 = v1_option if v1 is None else v1
            v2 = v2_option if v2 is None else v2

            if AIFF is not None and id3_filetype is AIFFFileType:
                if v2 == 3:
                    audio.tags.save(v2_version=3)  # AIFF doesn't support id3v1
                else:
                    audio.tags.save()  # AIFF doesn't support id3v1

            elif DSF is not None and id3_filetype is DSFFileType:
                if v2 == 3:
                    audio.tags.save(v2_version=3)  # DSF doesn't support id3v1
                else:
                    audio.tags.save()  # DSF doesn't support id3v1
            else:
                if v2 == 4:
                    audio.tags.update_to_v24()
                    audio.tags.save(v1=v1, v2_version=4)
                else:
                    c = ID3()
                    c.filename = self.filepath
                    c.update(audio)
                    c.update_to_v23()
                    c.save(v1=v1, v2_version=3)

            self.__tags['__tag_read'] = 'ID3v2.4' if v2 == 4 else 'ID3v2.3'
            self.update_tag_list()
            self._originaltags = list(audio.keys())

        def set_fundamentals(self, tags, mut_obj, images=None):
            self.__tags = tags
            self.mut_obj = mut_obj
            if images:
                self.images = images
            self._originaltags = list(tags.keys())
            self.set_attrs(ATTRIBUTES, tags)

        def to_encoding(self, encoding=UTF8):
            frames = []
            saved = []
            for frame in self.__tags.values():
                if hasattr(frame, 'encoding'):
                    frames.append((frame, frame.encoding))
                    frame.encoding = encoding
            try:
                self.save(v2=4)
            except:
                for frame, enc in frames:
                    frame.encoding = enc
                raise

        def update_tag_list(self):
            tag = self.__tags['__tag_read']
            l = tag_versions.tags_in_file(self.filepath)
            if l:
                if tag and tag in l:
                    l.remove(tag)
                    l.insert(0, tag)
                self.__tags['__tag'] = ', '.join(l)
            else:
                self.__tags['__tag'] = tag

    return Tag


Tag = tag_factory(ID3FileType)

filetypes = [
    (ID3FileType, tag_factory(ID3FileType), 'ID3', 'mp3'),
]

if AIFF is not None:
    filetypes.append((AIFFFileType, tag_factory(AIFFFileType), 'AIFF', 'aiff'))

if DSF is not None:
    filetypes.append((DSFFileType, tag_factory(DSFFileType), 'DSF', ['dsf', 'dff']))



# -*- coding: utf-8 -*-
from PyQt5.QtWidgets import (QCheckBox, QLabel, QHBoxLayout, QSpinBox,
                             QVBoxLayout, QWidget)

from .translations import translate


def sanitize(type_, value, default=None):
    if type_ is int:
        try:
            return int(value)
        except (TypeError, ValueError):
            return default
    elif type_ is bool:
        if value is True or value == 'True':
            return True
        else:
            return False
    else:
        return value


class AutoNumbering(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        def hbox(*widgets):
            box = QHBoxLayout()
            [box.addWidget(z) for z in widgets]
            box.addStretch()
            return box

        vbox = QVBoxLayout()

        startlabel = QLabel(translate('Autonumbering Wizard', "&Start: "))
        self._start = QSpinBox()
        startlabel.setBuddy(self._start)
        self._start.setValue(1)
        self._start.setMaximum(65536)

        vbox.addLayout(hbox(startlabel, self._start))

        label = QLabel(translate('Autonumbering Wizard', 'Max length after padding with zeroes: '))
        self._padlength = QSpinBox()
        label.setBuddy(self._padlength)
        self._padlength.setValue(1)
        self._padlength.setMaximum(65535)
        self._padlength.setMinimum(1)
        vbox.addLayout(hbox(label, self._padlength))

        self._restart_numbering = QCheckBox(translate('Autonumbering Wizard', "&Restart numbering at each directory."))

        vbox.addWidget(self._restart_numbering)
        vbox.addStretch()

        self.setLayout(vbox)

    def setArguments(self, *args):
        minimum = sanitize(int, args[0], 0)
        restart = sanitize(bool, args[1], True)
        padding = sanitize(int, args[2], 1)

        self._start.setValue(minimum)
        self._restart_numbering.setChecked(restart)
        self._padlength.setValue(padding)

    def arguments(self):
        x = [
            self._start.value(),
            self._restart_numbering.isChecked(),
            self._padlength.value()]
        return x


from .functions import autonumbering

dialogs = {autonumbering: AutoNumbering}



# -*- coding: utf-8 -*-
import tempfile

from PyQt5.QtCore import QRect, Qt
from PyQt5.QtGui import QFont, QFontMetrics, QPainter, QBrush
from PyQt5.QtSvg import QSvgGenerator
from PyQt5.QtWidgets import QHBoxLayout, QVBoxLayout, QWidget

from ..audioinfo.util import commonimages
from ..constants import KEEP, BLANK
from ..constants import RIGHTDOCK, SELECTIONCHANGED
from ..puddleobjects import PicWidget


def svg_to_pic(data, desc):
    return {'data': data, 'size': len(data),
            'description': desc, 'imagetype': 3}


def get_font(rect, *text):
    font = QFont()
    lengths = [(t, QFontMetrics(font).horizontalAdvance(t)) for t in text]
    lowest = max(lengths, key=lambda x: x[1])
    size = 12
    while QFontMetrics(font).horizontalAdvance(lowest[0]) < rect.width() - 30:
        font.setPointSize(size)
        size += 1
    return font


def create_svg(text, font, rect=None):
    if not rect:
        rect = QRect(0, 0, 200, 200)
    f = tempfile.NamedTemporaryFile()

    generator = QSvgGenerator()
    generator.setFileName(f.name)
    generator.setSize(rect.size())
    generator.setViewBox(rect)
    generator.setTitle("puddletag image")
    generator.setDescription("just to see")

    painter = QPainter()
    painter.begin(generator)
    painter.fillRect(rect, Qt.GlobalColor.black)
    painter.setFont(font)
    painter.setPen(Qt.GlobalColor.white)
    painter.setBrush(QBrush(Qt.GlobalColor.white))
    painter.drawText(rect, Qt.AlignmentFlag.AlignCenter, text)
    painter.end()

    svg = open(f.name).read()
    f.close()
    return svg


class ArtworkWidget(QWidget):
    def __init__(self, parent=None, status=None):
        QWidget.__init__(self, parent)
        self.receives = [(SELECTIONCHANGED, self.fill)]
        self.emits = []
        self.picwidget = PicWidget()
        vbox = QVBoxLayout()
        vbox.addWidget(self.picwidget)
        hbox = QHBoxLayout()
        hbox.addStrut(1)
        vbox.addLayout(hbox)
        self.setLayout(vbox)
        status['images'] = self.images
        self._audios = []
        self._status = status
        self._readOnly = None

    def fill(self, audios=None):
        if audios is None:
            audios = self._status['selectedfiles']
        if not audios:
            self.picwidget.setEnabled(False)
            self.picwidget.setImages(None)
            return
        if not self.isVisible():
            self._audios = audios
            return
        self.picwidget.currentFile = audios[0]
        self.picwidget.filePattern = self._status['cover_pattern']
        pics = list(self._readOnlyPics())
        self.picwidget.setEnabled(True)
        images = []
        imagetags = set()
        for audio in audios:
            imagetags = imagetags.union(audio.IMAGETAGS)
            if audio.IMAGETAGS:
                images.append(audio.images)
            else:
                images.append({})

        self.picwidget.lastfilename = audios[0].filepath
        images = commonimages(images)
        self.picwidget.setImageTags(imagetags)

        if images == 0:
            self.picwidget.setImages(pics, default=0)
            self.picwidget.context = 'Cover Varies'
        elif images is None:
            self.picwidget.setImages(pics, default=1)
        else:
            pics.extend(images)
            self.picwidget.setImages(pics, default=2)
        self.picwidget.readonly = [0, 1]

        if imagetags:
            self.setEnabled(True)
        else:
            self.setEnabled(False)

    def init(self):
        pics = self._readOnlyPics()
        self.picwidget.setImages(pics)
        self.picwidget.readonly = [0, 1]

    def _readOnlyPics(self):
        if not self._readOnly:
            font = get_font(QRect(0, 0, 200, 200), KEEP, BLANK)
            data = (create_svg(KEEP, font), create_svg(BLANK, font))
            self._readOnly = tuple([svg_to_pic(datum, desc) for datum, desc
                                    in zip(data, (KEEP, BLANK))])
        return self._readOnly

    def images(self):
        images = None
        if self.picwidget.currentImage == 1:  # <blank>
            images = []
        elif self.picwidget.currentImage > 1:  # <keep> is 0, so everything else.
            images = self.picwidget.images[2:]
        return images

    def showEvent(self, event):
        QWidget.showEvent(self, event)
        if self._audios:
            self.fill(self._audios)
            self._audios = []


control = ('Artwork', ArtworkWidget, RIGHTDOCK, False)



# -*- coding: utf-8 -*-

from os import path

version_string = '2.5.0'

# This is only used by the github workflow
_buildid = None
if _buildid:
    version_string += f".post{_buildid}"

changeset = None
if '__file__' in globals():
    filedir = path.dirname(__file__)
    git_root = path.join(filedir, '..')
    hash_file = path.join(git_root, '.git', 'HEAD')
    if path.exists(hash_file):
        try:
            with open(hash_file) as fo:
                changeset = fo.read().strip()
            if changeset.startswith("ref: "):
                hash_file = path.join(git_root, '.git', *(changeset[5:].split('/')))
                with open(hash_file) as fo:
                    changeset = fo.read().strip()
        except (EnvironmentError, AttributeError):
            pass

# for PEP-396 compatibility
__version__ = version_string



# -*- coding: utf-8 -*-
import os
import sys

from PyQt5.QtWidgets import QAction, QApplication, QMainWindow, QMenu, QMenuBar, QToolBar

from .constants import CONFIGDIR
from .puddleobjects import PuddleConfig, get_icon, open_resourcefile
from .translations import translate

__version__ = 33

files = [open_resourcefile(filename)
         for filename in ['data:./caseconversion.action', 'data:./standard.action']]

SEPARATOR = 'separator'
ALWAYS = 'always'
menu_path = os.path.join(CONFIGDIR, 'menus')
shortcut_path = os.path.join(CONFIGDIR, 'shortcuts')


def create_file(path, resource):
    text = open_resourcefile(resource).read()
    f = open(path, 'w')
    f.write(text)
    f.close()


def check_file(path: str, resource: str) -> None:
    """Check version of config file at path, and update with resource if needed."""
    if not os.path.exists(path):
        create_file(path, resource)
    else:
        cparser = PuddleConfig(path)
        version = cparser.get('info', 'version', 0)
        if version < __version__:
            print(f'Replacing version {version} config file {path} with newer version')
            create_file(path, resource)


def create_files():
    check_file(menu_path, 'data:./menus')
    check_file(shortcut_path, 'data:./shortcuts')


def get_menus(section, filepath=None):
    cparser = PuddleConfig()
    if not filepath:
        filepath = menu_path
    cparser.filename = filepath
    menus = []
    settings = cparser.data
    temp = settings[section]
    menus = [(z, temp[z]) for z in settings[section + 'attrs']['order']]
    return menus


def menubar(menus, actions):
    texts = [str(action.text()) for action in actions]

    menubar = QMenuBar()
    winmenu = None
    _menus = {}
    for title, actionlist in menus:
        menu = menubar.addMenu(translate("Menus", title))
        _menus[title] = [menu]
        if title == '&Windows':
            winmenu = menu
            tr_section = 'Dialogs'
        else:
            tr_section = 'Menus'
        for action in actionlist:
            if action in texts:
                shortcut = actions[texts.index(action)]
                shortcut.setText(translate(tr_section, action))
                menu.addAction(shortcut)
                _menus[title].append(shortcut)
            elif action == SEPARATOR:
                menu.addSeparator()
    return menubar, winmenu, _menus


def context_menu(section, actions, filepath=None):
    cparser = PuddleConfig(filepath)
    if not filepath:
        filepath = menu_path
        cparser.filename = filepath
    order = [translate('Menus', z) for z in cparser.get(section, 'order', [])]
    if not order:
        return
    texts = [str(action.text()) for action in actions]
    menu = QMenu()
    for action in order:
        if action in texts:
            menu.addAction(actions[texts.index(action)])
        elif action == SEPARATOR:
            menu.addSeparator()
    return menu


def toolbar(groups, actions, controls=None):
    texts = [str(action.text()) for action in actions]
    if controls:
        controls = dict([('widget-' + z, v) for z, v in controls.items()])
    toolbar = QToolBar('Toolbar')
    for name, actionlist in groups:
        for action in actionlist:
            if action in texts:
                toolbar.addAction(actions[texts.index(action)])
            elif action in controls:
                toolbar.addWidget(controls[action])
        toolbar.addSeparator()
    return toolbar


def create_action(win, name, control, command, icon=None, enabled=ALWAYS,
                  tooltip=None, shortcut=None, status=None, togglecheck=None,
                  checkstate=None, icon_name=None):
    if icon or icon_name:
        action = QAction(get_icon(icon_name, icon), name, win)
    else:
        action = QAction(name, win)
    action.setEnabled(False)

    if shortcut:
        try:
            action.setShortcut(shortcut)
        except TypeError:
            action.setShortcuts(shortcut)

    if tooltip:
        action.setToolTip(translate('Menus', tooltip))

    if togglecheck is not None:
        action.setCheckable(True)
        checked = int(checkstate)
        action.setChecked(bool(checked))

    action.togglecheck = togglecheck
    action.enabled = enabled
    action.command = command
    action.control = control
    action.status = status

    return action


def get_actions(parent, filepath=None):
    cparser = PuddleConfig()
    if not filepath:
        filepath = shortcut_path
    cparser.filename = filepath
    setting = cparser.data
    actions = []
    for section in cparser.sections():
        if section.startswith('shortcut'):
            values = dict([(str(k), v) for k, v in setting[section].items()])
            actions.append(create_action(parent, **values))
    return actions


if __name__ == '__main__':
    app = QApplication(sys.argv)
    win = QMainWindow()
    win.toolbar = win.addToolBar('toolbar')
    loadShortCuts()
    win.show()
    app.exec_()
